/*
 * File: search.c
 * Auth: Eric Harlow
 *
 * Modified: Andy Bell
 */

#include <gtk/gtk.h>
#include <string.h>
#include "dialog.h"
#include "desktop.h"

GtkWidget *dialog_window;
GtkWidget *entry;

static FindStringCallback fnFind;
static gchar *find_string;

static gboolean
file_selection_configure_event_cb(GtkWidget *widget, GdkEventConfigure *event,
				gpointer user_data);

void find_handler (GtkWidget *widget, gpointer data)
{
    if (find_string)
        g_free (find_string);

    find_string = gtk_editable_get_chars (GTK_EDITABLE (entry), 0, -1);

	if (!(*fnFind)(find_string, data))
		gtk_widget_destroy(dialog_window);
}


void cancel_handler (GtkWidget *widget, gpointer data)
{
	gtk_widget_destroy(dialog_window);
}


/*
 * FindStringDialog
 *
 * Function to display a search dialog
 */
void FindStringDialog (char *szMessage, FindStringCallback callback, gpointer data)
{
    GtkWidget *label;
    GtkWidget *button;
    GtkWidget *table;
	GtkWidget *hbuttonbox;

    dialog_window = gtk_dialog_new ();
    gtk_window_set_position(GTK_WINDOW(dialog_window),GTK_WIN_POS_CENTER);

    gtk_signal_connect (GTK_OBJECT (dialog_window), "destroy",
	                GTK_SIGNAL_FUNC (gtk_widget_destroy),
	                dialog_window);

    gtk_window_set_title (GTK_WINDOW (dialog_window), "Find");
    gtk_window_set_modal (GTK_WINDOW (dialog_window),TRUE);

    gtk_container_border_width (GTK_CONTAINER (dialog_window), 5);
    
	gtk_widget_set_usize(dialog_window, 240, 100);

	fnFind = callback;

    table = gtk_table_new(1, 2, FALSE);

    label = gtk_label_new (szMessage);
 
    entry = gtk_entry_new ();

    /* --- If they've searched before --- */
    if (find_string)
        /* --- Prefill with the value they looked for last time --- */
        gtk_entry_set_text (GTK_ENTRY (entry), find_string);

	gtk_misc_set_alignment(GTK_MISC(label), 1.0, 0.5);
	gtk_table_attach(GTK_TABLE(table), label, 0, 1, 0, 1, GTK_FILL, GTK_EXPAND, 0, 0);
	gtk_table_attach(GTK_TABLE(table), entry, 1, 2, 0, 1, GTK_FILL, GTK_EXPAND, 0, 0);

    gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog_window)->vbox), 
                        table, TRUE, TRUE, 0);

	hbuttonbox = gtk_hbutton_box_new();
	gtk_button_box_set_layout(GTK_BUTTON_BOX(hbuttonbox), GTK_BUTTONBOX_SPREAD);
	gtk_button_box_set_spacing(GTK_BUTTON_BOX(hbuttonbox), 4);

    button = gtk_button_new_with_label ("Find Next");
    gtk_signal_connect (GTK_OBJECT (button), "clicked",
						GTK_SIGNAL_FUNC (find_handler), data);

    gtk_box_pack_start (GTK_BOX (hbuttonbox), button, TRUE, TRUE, 0);

    button = gtk_button_new_with_label ("Close");

    gtk_signal_connect (GTK_OBJECT (button), "clicked",
	                GTK_SIGNAL_FUNC (cancel_handler), data);

    gtk_box_pack_start (GTK_BOX (hbuttonbox), button, TRUE, TRUE, 0);

    gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog_window)->action_area), 
			  hbuttonbox, TRUE, TRUE, 0);

    gtk_widget_show_all(dialog_window);
}

GtkWidget *FileSelectionDialog(const gchar *title)
{
	/* This overrides the default layout and appearance of GtkFileSelection
	 * so that it looks like other dialogs in EmLinux applications.
	 */
	GtkFileSelection *file_selection = GTK_FILE_SELECTION(gtk_file_selection_new(title));
	GtkWidget *confirm_area;
	
	/* --- Set the window title and size --- */
	gtk_widget_set_usize (GTK_WIDGET(file_selection), WINDOW_WIDTH, WINDOW_HEIGHT);
	gtk_window_set_policy(GTK_WINDOW(file_selection), TRUE,
				TRUE, TRUE);

	gtk_box_set_spacing(GTK_BOX(file_selection->main_vbox), 0);
	gtk_container_set_border_width(GTK_CONTAINER(file_selection), 4);

	confirm_area = file_selection->ok_button->parent;

	gtk_container_set_border_width(GTK_CONTAINER(confirm_area), 10);
	gtk_button_box_set_layout(GTK_BUTTON_BOX(confirm_area), GTK_BUTTONBOX_SPREAD);
	gtk_button_box_set_spacing(GTK_BUTTON_BOX(confirm_area), 4);
	GTK_WIDGET_UNSET_FLAGS (file_selection->ok_button, GTK_CAN_DEFAULT);
	GTK_WIDGET_UNSET_FLAGS (file_selection->cancel_button, GTK_CAN_DEFAULT);

	gtk_file_selection_hide_fileop_buttons(file_selection);
	gtk_signal_connect(GTK_OBJECT(file_selection), "configure_event",
	                   GTK_SIGNAL_FUNC(file_selection_configure_event_cb), NULL);

	return GTK_WIDGET(file_selection);
}


static gboolean
file_selection_configure_event_cb(GtkWidget *widget, GdkEventConfigure *event,
				gpointer user_data)
{
	gtk_widget_set_usize(widget, event->width, event->height);
	return TRUE;
}
