#include "scheduler.h"
#include <gtk/gtkdrawingarea.h>
#include <gtk/gtkinvisible.h>

#if (PCFB || PDA)
#include <sys/time.h>
#include <unistd.h>
#endif

#define FONT_NAME	"-adobe-helvetica-bold-r-normal--12-120-75-75-p-69-iso8859-1"

#define MONTH_X		14
#if (SCREEN_HEIGHT < 320)
#define MONTH_Y		5
#else
#define MONTH_Y		15
#endif
#define MONTH_WD	35
#if (SCREEN_HEIGHT < 320)
#define MONTH_HT	20
#else
#define MONTH_HT	22
#endif
#define MONTH_COLS	6
#define MONTH_ROWS	2

#define TOP_X		14
#if (SCREEN_HEIGHT < 320)
#define TOP_Y		70
#else
#define TOP_Y		90
#endif
#define NUM_COLS	7
#define NUM_ROWS	6
#define COL_WD		30
#if (SCREEN_HEIGHT < 320)
#define ROW_HT		19
#else
#define ROW_HT		22
#endif
#define DELTA_X		16
#define DELTA_Y		6

#define SPACING         4

#define LAST_YEAR	-1
#define THIS_YEAR	0
#define NEXT_YEAR	+1



static gchar *weekday[] = {"S", "M", "T", "W", "T", "F", "S"};
static GtkWidget *vbox;
static GtkWidget *label, *board;
static GdkGC *fg, *bg;
static GdkFont *font;
static task_info *tip;
static time_t tref;
static gint cur_m_col, cur_m_row;
static gint first_day, max_days, cur_d_col, cur_d_row, marks[32];

static gchar *monthnames[]={"",
			    "Jan", "Feb", "Mar", "Apr", "May", "Jun",
			    "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};



/******** External function declarations ********/

GtkWidget *get_toplevel_widget();
GtkWidget *get_layout_widget();
void show_main_window(gint flg);
void set_gc_black_white(GdkGC *fg, GdkGC *bg);
task_info *get_current_event();
void jump_to_date(time_t tref);
void set_current_time(task_info *tip);
int get_day_of_week(int year, int mm, int dd);
int get_days_in_month(int year, int mm);
int get_days_between(int y1, int m1, int d1, int y2, int m2, int d2);
gboolean get_monthly_summary(task_info *tip, gint *fs);
void get_grey_colors(GdkColor *grey1, GdkColor *grey2);
gint is_leap_year(int year);



/******** Internal function declarations ********/

static void make_topbox(GtkWidget *vbox);
static void make_showbox(GtkWidget *vbox);
static void draw_calendar(GtkWidget *widget, GdkEventExpose *event,
                          gpointer data);
static void draw_calendar_month(GdkWindow *wn);
static void draw_calendar_frame(GdkWindow *wn);
static void draw_calendar_days(GdkWindow *wn);
static void draw_month_block(GdkWindow *wn, GdkFont *fn, GdkGC *fg,
                             GdkGC *bg, gint x, gint y, gchar *buf);
static void draw_day_block(GdkWindow *wn, GdkFont *fn, GdkGC *fg,
                           GdkGC *bg, gint x, gint y, gchar *buf,
                           gint mark);
static gint month_selected(GtkWidget *widget, GdkEventButton *event,
                           gpointer data);
static gint day_selected(GtkWidget *widget, GdkEventButton *event,
                         gpointer data);
static void goto_selected(GtkWidget *widget, GdkEvent *event,
                          gpointer data);
static void goto_canceled(GtkWidget *widget, GdkEvent *event,
                          gpointer data);
static void make_month_display(GtkWidget *tbox);
static void change_year(GtkWidget *widget, gint mode);
static void button_pressed(GtkWidget *widget, GdkEventButton *event,
                           gpointer data);



/******** External function definitions ********/

void show_calendar_window(gint flg)
{
   if(flg)
      gtk_widget_map(vbox);
   else
      gtk_widget_unmap(vbox);
}



void make_calendar()
{
   if(vbox == NULL)
   {
      /* Create a vertical box for layout control */
      vbox = gtk_vbox_new(FALSE, ALL_SPACE);
      gtk_widget_show(vbox);
      gtk_fixed_put(GTK_FIXED(get_layout_widget()), vbox, 0, 0);

      make_topbox(vbox);

      make_showbox(vbox);

      // resize to current window size
      {
         GtkWidget *w = get_toplevel_widget();
         gtk_widget_set_usize( vbox, (w->allocation).width, (w->allocation).height );
      }
   }
   else
      show_calendar_window(1);

   show_main_window(0);
}



/******** Internal function definitions ********/

static void make_month_display(GtkWidget *tbox)
{
   gchar buf[40];

   tip = get_current_event();

   if(label == NULL)
   {
      /* Create a text label for showing the current date */
      tref = tip->tref;
      strftime(buf, sizeof buf, "%G", localtime(&tref));
      label = gtk_label_new(buf);
      gtk_box_pack_start(GTK_BOX(tbox), label, FALSE, FALSE, 5);
      gtk_widget_show(label);
   }
   else
   {
      /* Update the text label for the date */
      strftime(buf, sizeof buf, "%G", localtime(&tref));
      gtk_label_set_text(GTK_LABEL(label), buf);

      /* Redraw calendar */
      draw_calendar(board, NULL, NULL);
   }
}


static void make_topbox(GtkWidget *vbox)
{
   GtkWidget *tbox, *prev, *next;
   GtkWidget *button;


   /* Create a topbox for showing date information */
   tbox = gtk_hbox_new(FALSE, ALL_SPACE);
   gtk_widget_set_usize(GTK_WIDGET(tbox), TOPBOX_WD, TOPBOX_HT);
   gtk_box_pack_start(GTK_BOX(vbox), tbox, FALSE, FALSE, FIRST_SPACE);
   gtk_widget_show(tbox);

   /* Create two buttons for changing date */
   prev = gtk_button_new_with_label("<");
   next = gtk_button_new_with_label(">");

   gtk_box_pack_start(GTK_BOX(tbox), prev, FALSE, FALSE, 0);
   gtk_box_pack_start(GTK_BOX(tbox), next, FALSE, FALSE, 0);

   gtk_widget_show(prev);
   gtk_widget_show(next);

   /* Create a text label for showing the current date */
   make_month_display(tbox);

   gtk_signal_connect(GTK_OBJECT(prev), "clicked",
                      GTK_SIGNAL_FUNC(change_year),
                      (gpointer)LAST_YEAR);

   gtk_signal_connect(GTK_OBJECT(next), "clicked",
                      GTK_SIGNAL_FUNC(change_year),
                      (gpointer)NEXT_YEAR);

   button = gtk_button_new_with_label("Cancel");
   gtk_box_pack_end(GTK_BOX(tbox), button, FALSE, FALSE, 0);
   gtk_signal_connect(GTK_OBJECT(button), "clicked",
                      GTK_SIGNAL_FUNC(goto_canceled), NULL);
   gtk_widget_show(button);

   button = gtk_button_new_with_label("Go To");
   gtk_box_pack_end(GTK_BOX(tbox), button, FALSE, FALSE, 0);
   gtk_signal_connect(GTK_OBJECT(button), "clicked",
                      GTK_SIGNAL_FUNC(goto_selected), NULL);
   gtk_widget_show(button);
}


static guint size_allocate_handler_id;
static void size_allocate_handler(GtkWidget *widget, GtkAllocation *allocation, gpointer user_data)
{
	gtk_signal_disconnect(GTK_OBJECT(widget), size_allocate_handler_id);
//	gtk_widget_set_usize(widget, allocation->width, WINDOW_HEIGHT - allocation->y);
	gtk_drawing_area_size(GTK_DRAWING_AREA(widget), allocation->width, WINDOW_HEIGHT - allocation->y);
}



static void make_showbox(GtkWidget *vbox)
{
   board = gtk_drawing_area_new();

   gtk_box_pack_start(GTK_BOX(vbox), board, FALSE, FALSE, SECOND_SPACE);

   size_allocate_handler_id = gtk_signal_connect(GTK_OBJECT(board), "size-allocate",
   											     GTK_SIGNAL_FUNC(size_allocate_handler), NULL);

   gtk_widget_add_events(board, GDK_BUTTON_PRESS_MASK);

   gtk_signal_connect(GTK_OBJECT(board), "expose_event",
                      GTK_SIGNAL_FUNC(draw_calendar), NULL);

   gtk_signal_connect(GTK_OBJECT(board), "button_press_event",
                      GTK_SIGNAL_FUNC(button_pressed), NULL);

   gtk_widget_show(board);
}



static void draw_calendar(GtkWidget *widget, GdkEventExpose *event,
                          gpointer data)
{
   static GdkGC *gc;
   GdkColor grey1, grey2;
   GdkWindow *wn;

   wn = widget->window;

#if (PCFB || PDA)
   if(fg == NULL)
   {
      fg = gdk_gc_new(wn);
      bg = gdk_gc_new(wn);
      set_gc_black_white(fg, bg);
   }
#else
   fg = widget->style->black_gc;
   bg = widget->style->white_gc;
#endif

   /* Use default font to draw day numbers */
   font = widget->style->font;

   if(gc == NULL)
   {
      gc = gdk_gc_new(wn);
      get_grey_colors(&grey1, &grey2);
      gdk_gc_set_foreground(gc, &grey1);
   }

   gdk_draw_rectangle(wn, gc, TRUE, 0, 0, widget->allocation.width, widget->allocation.height);

   gdk_draw_line(wn, fg, 0, 0, widget->allocation.width, 0);
   gdk_draw_line(wn, bg, 0, 1, widget->allocation.width, 1);

   draw_calendar_month(wn);
   draw_calendar_frame(wn);
   draw_calendar_days(wn);
}



static void draw_calendar_month(GdkWindow *wn)
{
   gint i, j, k, w, h, x, y;

   w = MONTH_COLS * MONTH_WD;
   h = MONTH_ROWS * MONTH_HT;

   /* Fill calendar frame area white color */
   gdk_draw_rectangle(wn, bg, TRUE, MONTH_X, MONTH_Y, w, h);

   /* Draw calendar frame vertical lines */
   for(i = 0, x = MONTH_X, y = MONTH_Y; i <= MONTH_COLS; i++)
   {
      gdk_draw_line(wn, fg, x, y, x, y + h);
      x += MONTH_WD;
   }

   /* Draw calendar frame horizontal lines */
   for(j = 0, x = MONTH_X, y = MONTH_Y; j <= MONTH_ROWS; j++)
   {
      gdk_draw_line(wn, fg, x, y, x + w, y);
      y += MONTH_HT;
   }

   /* Draw month names in upper frame */
   for(j = 0, k = 1; j < MONTH_ROWS; j++)
   {
      for(i = 0; i < MONTH_COLS; i++, k++)
      {
         x = MONTH_X + i * MONTH_WD + 5;
         y = MONTH_Y + j * MONTH_HT + 15;

         if(k == tip->month)
         {
            draw_month_block(wn, font, bg, fg, x, y, monthnames[k]);

            cur_m_col = i;
            cur_m_row = j;
         }
         else
            gdk_draw_string(wn, font, fg, x + 4, y, monthnames[k]);
      }
   }
}



static void draw_calendar_frame(GdkWindow *wn)
{
   GdkFont *fn;
   gint i, j, w, h, x, y;

   w = NUM_COLS * COL_WD;
   h = NUM_ROWS * ROW_HT;

   /* Fill calendar frame area white color */
   gdk_draw_rectangle(wn, bg, TRUE, TOP_X, TOP_Y, w, h);

   /* Allocate font for weekday names */
   fn = gdk_font_load(FONT_NAME);

   /* Draw weekday names above frame */
   for(i = 0, x = TOP_X + 10, y = TOP_Y - 4; i < NUM_COLS; i++)
   {
      gdk_draw_string(wn, fn, fg, x, y, weekday[i]);
      x += COL_WD;
   }

   /* Deallocate font structure loaded */
   gdk_font_unref(fn);

   /* Draw calendar frame vertical lines */
   for(i = 0, x = TOP_X, y = TOP_Y; i <= NUM_COLS; i++)
   {
      gdk_draw_line(wn, fg, x, y, x, y + h);
      x += COL_WD;
   }

   /* Draw calendar frame horizontal lines */
   for(j = 0, x = TOP_X, y = TOP_Y; j <= NUM_ROWS; j++)
   {
      gdk_draw_line(wn, fg, x, y, x + w, y);
      y += ROW_HT;
   } 
}



static void draw_calendar_days(GdkWindow *wn)
{
   gchar buf[40];
   gint a, d, i, j, k, m, n, x, y;

   /* Use local reference time to draw calendar */
   strftime(buf, sizeof buf, "%Y %m %e", localtime(&tref));
   sscanf(buf, "%d %d %d", &y, &m, &d);

   k = get_day_of_week(y, m, 1);
   n = get_days_in_month(y, m);
   a = 1;

   first_day = k;
   max_days = n;

   get_monthly_summary(tip, marks);

   /* Draw day numbers in frame */
   for(j = 0, y = TOP_Y + 15; j < NUM_ROWS; j++, k = 0)
   {
      for(i = 0, x = TOP_X + 5; i < NUM_COLS; i++)
      {
         if(i >= k)
         {
            sprintf(buf, "%d", a);

            if(a == d)
            {
               /* Highlight the current day */
               draw_day_block(wn, font, bg, fg, x, y, buf, marks[a]);

               cur_d_col = i;
               cur_d_row = j;
            }
            else
            {
               gdk_draw_string(wn, font, fg, x, y, buf);

               if(marks[a])
                  gdk_draw_string(wn, font, fg, x + DELTA_X, y + DELTA_Y, "+");
            }

            if(a++ >= n)
               return;
         }

         x += COL_WD;
      }

      y += ROW_HT;
   }
}




static gint month_selected(GtkWidget *widget, GdkEventButton *event,
                           gpointer data)
{
   GdkWindow *wn;
   gint d1, d2, dd, m1, m2, x, y, col, row;

   col = (event->x - MONTH_X) / MONTH_WD;
   row = (event->y - MONTH_Y) / MONTH_HT;

   m1 = cur_m_col + cur_m_row * MONTH_COLS + 1;
   m2 = col + row * MONTH_COLS + 1;

   if(m2 < 1 || m2 > 12)
      return -1;

   if(m1 == m2 || col < 0 || col >= MONTH_COLS || row >= MONTH_ROWS)
      return 0;

   wn = widget->window;

   /* draw the last selected month block in normal */
   x = MONTH_X + cur_m_col * MONTH_WD + 5;
   y = MONTH_Y + cur_m_row * MONTH_HT + 15;

   draw_month_block(wn, font, fg, bg, x, y, monthnames[m1]);

   /* draw the current selected month block in reverse */
   x = MONTH_X + col * MONTH_WD + 5;
   y = MONTH_Y + row * MONTH_HT + 15;

   draw_month_block(wn, font, bg, fg, x, y, monthnames[m2]);

   /* save row and column index for the current month */
   cur_m_col = col;
   cur_m_row = row;

   /* get current day and check valid day for the change */
   d1 = cur_d_col + cur_d_row * 7 - first_day + 1;
   dd = get_days_in_month(tip->year, m2);

   if(d1 > dd)
      d2 = dd;
   else
      d2 = d1;

   dd = get_days_between(tip->year, m1, d1, tip->year, m2, d2);

   tref += dd * 24 * 3600;

   /* update calendar display */
   tip->tref = tref;
   set_current_time(tip);
   get_monthly_summary(tip, marks);

   make_month_display(NULL);

   return 1;
}



static gint day_selected(GtkWidget *widget, GdkEventButton *event,
                         gpointer data)
{
   GdkWindow *wn;
   gchar buf[40];
   gint d1, d2, x, y, col, row;

   col = (event->x - TOP_X) / COL_WD;
   row = (event->y - TOP_Y) / ROW_HT;

   d1 = cur_d_col + cur_d_row * NUM_COLS - first_day + 1;
   d2 = col + row * NUM_COLS - first_day + 1;

   if(d2 < 1 || d2 > max_days)
      return -1;

   if(d1 == d2 || col < 0 || col >= NUM_COLS || row >= NUM_ROWS)
      return 0;

   wn = widget->window;

   /* draw the last selected day block in normal */
   x = TOP_X + cur_d_col * COL_WD + 5;
   y = TOP_Y + cur_d_row * ROW_HT + 15;

   sprintf(buf, "%d", d1);
   draw_day_block(wn, font, fg, bg, x, y, buf, marks[d1]);

   /* draw the current selected day block in reverse */
   x = TOP_X + col * COL_WD + 5;
   y = TOP_Y + row * ROW_HT + 15;

   sprintf(buf, "%d", d2);
   draw_day_block(wn, font, bg, fg, x, y, buf, marks[d2]);

   /* save row and column index for the current day */
   cur_d_col = col;
   cur_d_row = row;

   /* update time record here */
   tref += (d2 - d1) * 24 * 3600;

   return 1;
}



static void draw_month_block(GdkWindow *wn, GdkFont *fn, GdkGC *fg,
                             GdkGC *bg, gint x, gint y, gchar *buf)
{
   gdk_draw_rectangle(wn, bg, TRUE, x - 5 + 1, y - 15 + 1,
                      MONTH_WD - 1, MONTH_HT - 1);

   gdk_draw_string(wn, fn, fg, x + 3, y, buf);
}



static void draw_day_block(GdkWindow *wn, GdkFont *fn, GdkGC *fg,
                           GdkGC *bg, gint x, gint y, gchar *buf,
                           gint mark)
{
   gdk_draw_rectangle(wn, bg, TRUE, x - 5 + 1, y - 15 + 1,
                      COL_WD - 1, ROW_HT - 1);

   gdk_draw_string(wn, fn, fg, x, y, buf);

   if(mark)
      gdk_draw_string(wn, fn, fg, x + DELTA_X, y + DELTA_Y, "+");
}



static void goto_selected(GtkWidget *widget, GdkEvent *event, gpointer data)
{
   show_main_window(1);

   show_calendar_window(0);

   jump_to_date(tref);
}


static void goto_canceled(GtkWidget *widget, GdkEvent *event, gpointer data)
{
   show_main_window(1);

   show_calendar_window(0);
}



static void change_year(GtkWidget *widget, gint mode)
{
   gchar buf[40];
   gint dd, y1, m1, d1, y2, m2, d2;
   gint offset;

   switch(mode)
   {
      case LAST_YEAR: offset = -1; break;
      case THIS_YEAR: offset =  0; break;
      case NEXT_YEAR: offset = +1; break;
      default: return;
   }

   /* get current day and check valid day for the change */
   strftime(buf, sizeof buf, "%Y %m %e", localtime(&tref));
   sscanf(buf, "%d %d %d", &y1, &m1, &d1);

   y2 = y1 + offset;
   m2 = m1;

   /* these are the beginning and end of the day! */
   if(y2 < 1970 || y2 > 2037)
      return;

   /* Try to keep the day unchanged */
   dd = get_days_in_month(y2, m2);

   if(d1 > dd)
      d2 = dd;
   else
      d2 = d1;

   /* Update the local reference time */
   dd = get_days_between(y1, m1, d1, y2, m2, d2);

   tref += dd * 24 * 3600;

   /* update calendar display */
   tip->tref = tref;
   set_current_time(tip);
   get_monthly_summary(tip, marks);

   make_month_display(NULL);
}



static void button_pressed(GtkWidget *widget, GdkEventButton *event,
                           gpointer data)
{
   gint col, row;

   col = cur_d_col;
   row = cur_d_row;

   if(month_selected(widget, event, data) >= 0)
   {
      /* do nothing here !!! */
   }
   else if(day_selected(widget, event, data) >= 0)
   {
      if(col == cur_d_col && row == cur_d_row)
         goto_selected(NULL, NULL, NULL);
   }
}



void
scheduler_resize_calendar( gint width, gint height ) {
   if (vbox) gtk_widget_set_usize( vbox, width, height );
   return;
}
