#include "scheduler.h"


#define USE_ALL_FUNCS


/***************************************************************************/
/* The following date routines are taken from the lib_date package.  Keep
 * them seperate in case we want to update them if a newer lib_date comes
 * out with fixes.  */

typedef  unsigned   int     N_int;
typedef  unsigned   long    N_long;
typedef  signed     long    Z_long;
typedef enum { false = FALSE , true = TRUE } boolean;

#define and         &&      /* logical (boolean) operators: lower case */
#define or          ||

static const N_int month_length[2][13] =
{
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 }
};

static const N_int days_in_months[2][14] =
{
  { 0, 0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334, 365 },
  { 0, 0, 31, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335, 366 }
};

static Z_long  calc_days(N_int year, N_int mm, N_int dd);
static N_int   day_of_week(N_int year, N_int mm, N_int dd);
static Z_long  dates_difference(N_int year1, N_int mm1, N_int dd1,
                                N_int year2, N_int mm2, N_int dd2);
static N_int   weeks_in_year(N_int year);

static boolean
leap(N_int year)
{
  return((((year % 4) == 0) and ((year % 100) != 0)) or ((year % 400) == 0));
}

static N_int
day_of_week(N_int year, N_int mm, N_int dd)
{
  Z_long  days;

  days = calc_days(year, mm, dd);
  if (days > 0L)
    {
      days--;
      days %= 7L;
      days++;
    }
  return( (N_int) days );
}

static N_int weeks_in_year(N_int year)
{
  return(52 + ((day_of_week(year,1,1)==4) or (day_of_week(year,12,31)==4)));
}


static boolean
check_date(N_int year, N_int mm, N_int dd)
{
  if (year < 1) return(false);
  if ((mm < 1) or (mm > 12)) return(false);
  if ((dd < 1) or (dd > month_length[leap(year)][mm])) return(false);
  return(true);
}

static N_int
week_number(N_int year, N_int mm, N_int dd)
{
  N_int first;

  first = day_of_week(year,1,1) - 1;
  return( (N_int) ( (dates_difference(year,1,1, year,mm,dd) + first) / 7L ) +
          (first < 4) );
}

static Z_long
year_to_days(N_int year)
{
  return( year * 365L + (year / 4) - (year / 100) + (year / 400) );
}


static Z_long
calc_days(N_int year, N_int mm, N_int dd)
{
  boolean lp;

  if (year < 1) return(0L);
  if ((mm < 1) or (mm > 12)) return(0L);
  if ((dd < 1) or (dd > month_length[(lp = leap(year))][mm])) return(0L);
  return( year_to_days(--year) + days_in_months[lp][mm] + dd );
}

#ifdef USE_ALL_FUNCS
static boolean
week_of_year(N_int *week, N_int *year, N_int mm, N_int dd)
{
  if (check_date(*year,mm,dd))
    {
      *week = week_number(*year,mm,dd);
      if (*week == 0)
        *week = weeks_in_year(--(*year));
      else if (*week > weeks_in_year(*year))
        {
          *week = 1;
          (*year)++;
        }
      return(true);
    }
  return(false);
}
#endif

static Z_long
dates_difference(N_int year1, N_int mm1, N_int dd1,
                 N_int year2, N_int mm2, N_int dd2)
{
  return( calc_days(year2, mm2, dd2) - calc_days(year1, mm1, dd1) );
}

/** END OF lib_date routines ************************************************/


int get_day_of_week(int year, int mm, int dd)
{
   int n;

   n = day_of_week((N_int)year, (N_int)mm, (N_int)dd);

   if(n >= 7)
      n -= 7;

   return n;
}

int get_days_in_month(int year, int mm)
{
  boolean lp;

  if (year < 1) return(0L);
  if ((mm < 1) or (mm > 12)) return(0L);
  return month_length[(lp = leap(year))][mm];
}

glong get_days_between(int y1, int m1, int d1, int y2, int m2, int d2)
{
   return dates_difference((N_int)y1, (N_int)m1, (N_int)d1,
                           (N_int)y2, (N_int)m2, (N_int)d2);
}

glong calculate_days(int year, int mm, int dd)
{
   return calc_days((N_int)year, (N_int)mm, (N_int)dd);
}

gint is_leap_year(int year)
{
   return leap(year);
}

