#include "scheduler.h"
#include <gtk/gtkinvisible.h>
#include <gtk/gtktext.h>



#define BOARD_WD        210
#define BOARD_HT        180

#define SPACING         4

#define LAST_HOUR	-1
#define THIS_HOUR	0
#define NEXT_HOUR	+1

#define DEF_SUMMARY	"No summary"



static GtkWidget *vbox;
static GtkWidget *label, *entry, *text;
static task_info *tip;

static guint xpadding = 0;
static guint ypadding = 2;


/******** External function declarations ********/

GtkWidget *get_toplevel_widget();
GtkWidget *get_layout_widget();
void show_main_window(gint flg);
task_info *get_current_event();
void set_current_time(task_info *tip);
gboolean read_event_details(task_info *tip);
gboolean save_event_details(task_info *tip);



/******** Internal function declarations ********/

static void make_topbox(GtkWidget *vbox);
static void make_showbox(GtkWidget *vbox);
static void save_if_needed();
static void done_selected(GtkWidget *widget, GdkEvent *event, gpointer data);
static void clear_selected(GtkWidget *widget, GdkEvent *event, gpointer data);
static void cancel_selected(GtkWidget *widget, GdkEvent *event, gpointer data);
static void make_time_display(GtkWidget *tbox);
static void change_hour(GtkWidget *widget, gint mode);
static void set_schedule_display();



/******** External function definitions ********/

void show_details_window(gint flg)
{
   if(flg)
      gtk_widget_map(vbox);
   else
      gtk_widget_unmap(vbox);
}



void make_event_details()
{
   if(vbox == NULL)
   {
      /* Create a vertical box for layout control */
      vbox = gtk_vbox_new(FALSE, ALL_SPACE);
      gtk_widget_show(vbox);
      gtk_fixed_put(GTK_FIXED(get_layout_widget()), vbox, 0, 0);

      /* create date display area and change hour buttons */
      make_topbox(vbox);

      /* create schedule list */
      make_showbox(vbox);

      // resize to current window size
      {
         GtkWidget *w = get_toplevel_widget();
         gtk_widget_set_usize( vbox, (w->allocation).width, (w->allocation).height );
      }
   }
   else
   {
      show_details_window(1);

      /* just to refresh date display */
      make_time_display(NULL);
   }

   show_main_window(0);
}



/******** Internal function definitions ********/

static void make_time_display(GtkWidget *tbox)
{
   gchar buf[40];

   tip = get_current_event();

   strftime(buf, sizeof buf, "%k:%M, %a %e %b %G", localtime(&tip->tref));

   if(label == NULL)
   {
      /* Create a text label for showing the current date */
      label = gtk_label_new(buf);
      gtk_box_pack_start(GTK_BOX(tbox), label, FALSE, FALSE, 5);
      gtk_widget_show(label);
   }
   else
   {
      /* Update the text label for the date */
      gtk_label_set_text(GTK_LABEL(label), buf);

      /* get current schedule and display it */
      read_event_details(tip);
      set_schedule_display();
   }
}



static void make_topbox(GtkWidget *vbox)
{
   GtkWidget *tbox, *prev, *next;
   GtkWidget *button;

   /* Create a topbox for showing date information */
   tbox = gtk_hbox_new(FALSE, ALL_SPACE);
   gtk_widget_set_usize(GTK_WIDGET(tbox), TOPBOX_WD, TOPBOX_HT);
   gtk_box_pack_start(GTK_BOX(vbox), tbox, FALSE, FALSE, FIRST_SPACE);
   gtk_widget_show(tbox);

   /* Create two buttons for changing date, a dummy as place holder */
   prev = gtk_button_new_with_label("<");
   next = gtk_button_new_with_label(">");

   gtk_box_pack_start(GTK_BOX(tbox), prev, FALSE, FALSE, 0);
   gtk_box_pack_start(GTK_BOX(tbox), next, FALSE, FALSE, 0);

   /* Create a text label for showing the current date */
   make_time_display(tbox);

   gtk_widget_show(prev);
   gtk_widget_show(next);

   gtk_signal_connect(GTK_OBJECT(prev), "clicked",
                      GTK_SIGNAL_FUNC(change_hour),
                      (gpointer)LAST_HOUR);

   gtk_signal_connect(GTK_OBJECT(next), "clicked",
                      GTK_SIGNAL_FUNC(change_hour),
                      (gpointer)NEXT_HOUR);


   button = gtk_button_new_with_label("Cancel");
   gtk_box_pack_end(GTK_BOX(tbox), button, FALSE, FALSE, 0);
   gtk_signal_connect(GTK_OBJECT(button), "clicked",
                      GTK_SIGNAL_FUNC(cancel_selected), NULL);
   gtk_widget_show(button);

   button = gtk_button_new_with_label("Delete");
   gtk_box_pack_end(GTK_BOX(tbox), button, FALSE, FALSE, 0);
   gtk_signal_connect(GTK_OBJECT(button), "clicked",
                      GTK_SIGNAL_FUNC(clear_selected), NULL);
   gtk_widget_show(button);

   button = gtk_button_new_with_label("Save");
   gtk_box_pack_end(GTK_BOX(tbox), button, FALSE, FALSE, 0);
   gtk_signal_connect(GTK_OBJECT(button), "clicked",
                      GTK_SIGNAL_FUNC(done_selected), NULL);
   gtk_widget_show(button);
}


guint text_size_allocate_handler_id;
void text_size_allocate_handler(GtkWidget *widget, GtkAllocation *allocation, gpointer user_data)
{
	gtk_signal_disconnect(GTK_OBJECT(widget), text_size_allocate_handler_id);
	gtk_widget_set_usize(widget, allocation->width, WINDOW_HEIGHT - allocation->y - ypadding);
}


static void make_showbox(GtkWidget *vbox)
{
   GtkWidget *table, *label, *vscroll;

   table = gtk_table_new(2, 3, FALSE);
   gtk_box_pack_start(GTK_BOX(vbox), table, TRUE, TRUE, 0);

   label = gtk_label_new("Summary:");
   gtk_misc_set_alignment(GTK_MISC(label), 1.0, 0.5);
   gtk_table_attach(GTK_TABLE(table), label, 0, 1, 0, 1,
                    //GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND,
                    0, 0,
                    xpadding, ypadding);

   entry = gtk_entry_new();
   gtk_table_attach(GTK_TABLE(table), entry, 1, 2, 0, 1,
                    //GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND,
                    GTK_FILL | GTK_SHRINK | GTK_EXPAND, 0,
                    xpadding, ypadding);

   label = gtk_label_new("Details:");
   gtk_misc_set_alignment(GTK_MISC(label), 1.0, 0.0);
   gtk_table_attach(GTK_TABLE(table), label, 0, 1, 1, 2,
                    //GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND,
                    0, GTK_FILL | GTK_SHRINK | GTK_EXPAND,
                    xpadding, ypadding);

   text = gtk_text_new(NULL, NULL);
   gtk_text_set_editable(GTK_TEXT(text), TRUE);
   gtk_table_attach(GTK_TABLE(table), text, 1, 2, 1, 2,
                    //GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND,
                    GTK_FILL | GTK_SHRINK | GTK_EXPAND, GTK_FILL | GTK_SHRINK | GTK_EXPAND,
                    xpadding, ypadding);
   
   vscroll = gtk_vscrollbar_new(GTK_TEXT(text)->vadj);
   gtk_table_attach(GTK_TABLE(table), vscroll, 2, 3, 1, 2,
                    //GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND,
                    GTK_FILL , GTK_FILL | GTK_SHRINK | GTK_EXPAND,
                    xpadding, ypadding);

   gtk_widget_show_all(vbox);
   //gtk_widget_show(vbox);

   text_size_allocate_handler_id = gtk_signal_connect(GTK_OBJECT(text), "size-allocate",
                                                      GTK_SIGNAL_FUNC(text_size_allocate_handler), NULL);

   set_schedule_display();
}



static void save_if_needed()
{
   gchar *s, *d;
   gint doit;

   if(tip != NULL)
   {
      s = gtk_entry_get_text(GTK_ENTRY(entry));
      d = gtk_editable_get_chars(GTK_EDITABLE(text), 0, -1);

      doit = 0;

      /* update summary if needed */
      if(strcmp(s, tip->summary) != 0)
      {
         g_free(tip->summary);
         tip->summary = g_strdup(s);
         doit = 1;
      }

      /* update details if needed */
      if(strcmp(d, tip->details) != 0)
      {
         g_free(tip->details);
         tip->details = g_strdup(d);
         doit = 1;
      }

      /* make a default summary if details is added */
      if(strcmp(tip->summary, "") == 0 && strcmp(tip->details, "") != 0)
      {
         g_free(tip->summary);
         tip->summary = g_strdup(DEF_SUMMARY);
         doit = 1;
      }

      /* save to file if needed */
      if(doit)
         save_event_details(tip);
   }

}



static void done_selected(GtkWidget *widget, GdkEvent *event, gpointer data)
{
   save_if_needed();

   show_main_window(1);

   show_details_window(0);
}



static void clear_selected(GtkWidget *widget, GdkEvent *event, gpointer data)
{
   /* Display text for schedule summary in entry box */
   gtk_entry_set_text(GTK_ENTRY(entry), "");

   /* Delete old text for schedule details */
   gtk_editable_delete_text(GTK_EDITABLE(text), 0, -1);

   done_selected(NULL, NULL, NULL);
}


static void cancel_selected(GtkWidget *widget, GdkEvent *event, gpointer data)
{
	show_main_window(1);

	show_details_window(0);
}



static void change_hour(GtkWidget *widget, gint mode)
{
   gint offset;

   save_if_needed();

   switch(mode)
   {
      case LAST_HOUR: offset = -1; break;
      case THIS_HOUR: offset =  0; break;
      case NEXT_HOUR: offset = +1; break;
      default: return;
   }

   tip = get_current_event();

   tip->row  += offset;
   tip->tref += offset * 3600;

   set_current_time(tip);

   make_time_display(NULL);
}



static void set_schedule_display()
{
   gint n;

   /* Display text for schedule summary in entry box */
   gtk_entry_set_text(GTK_ENTRY(entry), tip->summary);

   /* Delete old and then insert new text for schedule details */
   gtk_editable_delete_text(GTK_EDITABLE(text), 0, -1);

   n = 0;

   gtk_editable_insert_text(GTK_EDITABLE(text),
                            tip->details, strlen(tip->details), &n);
}


void scheduler_resize_details( gint width, gint height ) {
   if (vbox) gtk_widget_set_usize( vbox, width, height );
   return;
}
