#include "scheduler.h"
#include <stdlib.h>
#include <gtk/gtkinvisible.h>


#define YESTERDAY       -1
#define TODAY           0
#define TOMORROW        +1


static GtkWidget *fixed, *clist;
static GtkWidget *sclwin;
static task_info info;



/******** External function declarations ********/

//void clear_entry_buffer();
//void close_summary_entry();
void clist_set_function(GtkWidget *clist);
int calculate_days(int year, int mm, int dd);
void get_grey_colors(GdkColor *darkgrey, GdkColor *lightgrey);
task_info *get_current_event();
gboolean read_event_details(task_info *tip);
gboolean save_event_details(task_info *tip);
void read_task_for_day(GtkWidget *clist, task_info *info);
gint save_current_task(task_info *info);
gboolean read_single_record(task_info *tip);
gboolean save_single_record(task_info *tip);
static void edit_details(GtkWidget *widget, gpointer gdata);
static void select_date(GtkWidget *widget, gpointer gdata);
void make_event_details();
void make_calendar();



/******** Internal function declarations ********/

static void make_date_display(GtkWidget *tbox);
static void change_day(GtkWidget *widget, gint mode);
static void init_list_items();



/******** External function definitions ********/

void set_current_time(task_info *tip)
{
   gchar buf[40];

   strftime(buf, sizeof buf, "%Y %m %e %k %M", localtime(&tip->tref));

   sscanf(buf, "%d %d %d %d %d", &tip->year, &tip->month, &tip->day,
          &tip->hour, &tip->minute);

   /* fixe me: should set row index appropriately */

   if(tip->row < 0)
      tip->row = 0;

   if(tip->row > NUM_HOURS)
      tip->row = NUM_HOURS;
}



task_info *get_current_event()
{
   return &info;
}



void jump_to_date(time_t tref)
{
   info.tref = tref;

   set_current_time(&info);
   make_date_display(NULL);
}



void fixed_put_child(GtkWidget *child, gint x, gint y)
{
   gtk_fixed_put(GTK_FIXED(fixed), child, x, y);
}



void fixed_move_child(GtkWidget *child, gint x, gint y)
{
   gtk_fixed_move(GTK_FIXED(fixed), child, x, y);
}



void save_schedule()
{
   gchar *s;

   gtk_clist_get_text(GTK_CLIST(clist), info.row, 0, &s);

   sscanf(s, "%d", &info.hour);
   sscanf(s + 3, "%d", &info.minute);

   gtk_clist_get_text(GTK_CLIST(clist), info.row, 1, &s);
   g_free(info.summary);
   info.summary = g_strdup(s);

   save_current_task(&info);
}



gboolean read_event_details(task_info *tip)
{
   gboolean f;

   if(clist)
   {
//      gtk_clist_get_text(GTK_CLIST(clist), tip->row, 1, &tip->summary);

      /* read schedule details from file */

      f = read_single_record(tip);
   }
   else
      f = FALSE;

   return f;
}



gboolean save_event_details(task_info *tip)
{
   gboolean f;

   if(clist && tip->row >= 0)
   {
      gtk_clist_set_text(GTK_CLIST(clist), tip->row, 1, tip->summary);

      /* save schedule details to file */

      f = save_single_record(tip);
   }
   else
      f = FALSE;

   return f;
}



void make_event_date(GtkWidget *vbox)
{
   GtkWidget *tbox, *prev, *next;//, *dummy;
   GtkWidget *button;

   /* Create a topbox for showing date information */
   tbox = gtk_hbox_new(FALSE, ALL_SPACE);
   gtk_widget_set_usize(GTK_WIDGET(tbox), TOPBOX_WD, TOPBOX_HT);
   gtk_box_pack_start(GTK_BOX(vbox), tbox, FALSE, FALSE, FIRST_SPACE);
   gtk_widget_show(tbox);

   /* Create two buttons for changing date */
   prev = gtk_button_new_with_label("<");
   next = gtk_button_new_with_label(">");

   gtk_box_pack_start(GTK_BOX(tbox), prev, FALSE, FALSE, 0);
   gtk_box_pack_start(GTK_BOX(tbox), next, FALSE, FALSE, 0);
   
   /* Create a text label for showing the current date */
   make_date_display(tbox);

   button = gtk_button_new_with_label("Go To...");
   gtk_box_pack_end(GTK_BOX(tbox), button, FALSE, FALSE, 0);
   gtk_widget_show(button);

   gtk_signal_connect(GTK_OBJECT(button), "clicked",
                      GTK_SIGNAL_FUNC(select_date), NULL);
   
   button = gtk_button_new_with_label("Edit");
   gtk_box_pack_end(GTK_BOX(tbox), button, FALSE, FALSE, 0);
   gtk_widget_show(button);

   gtk_signal_connect(GTK_OBJECT(button), "clicked",
                      GTK_SIGNAL_FUNC(edit_details), NULL);
   
   gtk_widget_show(prev);
   gtk_widget_show(next);

   gtk_signal_connect(GTK_OBJECT(prev), "clicked",
                      GTK_SIGNAL_FUNC(change_day), (gpointer)YESTERDAY);
   gtk_signal_connect(GTK_OBJECT(next), "clicked",
                      GTK_SIGNAL_FUNC(change_day), (gpointer)TOMORROW);
}


guint list_size_allocate_handler_id;
void list_size_allocate_handler(GtkWidget *widget, GtkAllocation *allocation, gpointer user_data)
{
	gtk_signal_disconnect(GTK_OBJECT(widget), list_size_allocate_handler_id);
	gtk_widget_set_usize(widget, allocation->width, WINDOW_HEIGHT - allocation->y);
}



void make_event_list(GtkWidget *vbox)
{
   gint offset = 2;

   sclwin = gtk_scrolled_window_new(NULL, NULL);
   gtk_container_border_width(GTK_CONTAINER(sclwin), offset);
   gtk_box_pack_start(GTK_BOX(vbox), sclwin, TRUE, TRUE, SECOND_SPACE);
   gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW (sclwin),
                        GTK_POLICY_NEVER, GTK_POLICY_ALWAYS);
   list_size_allocate_handler_id = gtk_signal_connect(GTK_OBJECT(sclwin), "size-allocate",
													  GTK_SIGNAL_FUNC(list_size_allocate_handler), NULL);
   gtk_widget_show(sclwin);

   /* Create a clist with 2 columns */
   clist = gtk_clist_new(2);
   gtk_container_border_width(GTK_CONTAINER(clist), 0);
   gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW (sclwin), clist);

   gtk_widget_show(clist);

   /* Set the column widths */
   gtk_clist_set_column_width(GTK_CLIST(clist), TIME_INX, TIME_WD);
   gtk_clist_set_column_width(GTK_CLIST(clist), EVENT_INX, EVENT_WD);
   
   /* Set the row height */
   gtk_clist_set_row_height(GTK_CLIST(clist), LIST_ROW_HT);

   /* Set text justification on each of the columns */
   gtk_clist_set_column_justification(GTK_CLIST(clist), 0, GTK_JUSTIFY_RIGHT);
   gtk_clist_set_column_justification(GTK_CLIST(clist), 1, GTK_JUSTIFY_LEFT);

   clist_set_function(clist);

   init_list_items();

   set_current_time(&info);

   read_task_for_day(clist, &info);

   /* Position the normal working hour rows in the middle */
/* this doesn't work !!!!
   gtk_clist_thaw(GTK_CLIST(clist));
   gtk_clist_moveto(GTK_CLIST(clist), (gint)(20), 0, 0.5, 0);
*/
}

#if 0
void make_event_list(GtkWidget *vbox)
{
   gint offset;

   offset = 2;

   sclwin = gtk_scrolled_window_new(NULL, NULL);
   gtk_container_border_width(GTK_CONTAINER(sclwin), -offset);
   gtk_widget_set_usize(GTK_WIDGET(sclwin), SWLIST_WD + offset * 4, SWLIST_HT);
   gtk_box_pack_start(GTK_BOX(vbox), sclwin, FALSE, FALSE, SECOND_SPACE);
   gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW (sclwin),
                        GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
   gtk_widget_show(sclwin);

   fixed = gtk_fixed_new();
   gtk_container_border_width(GTK_CONTAINER(fixed), -offset);
   gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW (sclwin), fixed);
   gtk_widget_show(fixed);

   /* Create a clist with 2 columns */
   clist = gtk_clist_new(2);
   gtk_container_border_width(GTK_CONTAINER(clist), 0);
   gtk_widget_set_usize(GTK_WIDGET(clist), SWLIST_WD + offset * 4,
                        (LIST_ROW_HT + 1) * NUM_HOURS + 2);
   gtk_fixed_put(GTK_FIXED(fixed), clist, 0, 0);
   gtk_widget_show(clist);

   /* Set the column widths */
   gtk_clist_set_column_width(GTK_CLIST(clist), TIME_INX, TIME_WD);
   gtk_clist_set_column_width(GTK_CLIST(clist), EVENT_INX, EVENT_WD);
   
   /* Set the row height */
   gtk_clist_set_row_height(GTK_CLIST(clist), LIST_ROW_HT);

   /* Set text justification on each of the columns */
   gtk_clist_set_column_justification(GTK_CLIST(clist), 0, GTK_JUSTIFY_RIGHT);
   gtk_clist_set_column_justification(GTK_CLIST(clist), 1, GTK_JUSTIFY_LEFT);

   clist_set_function(clist);

   init_list_items();

   set_current_time(&info);

   read_task_for_day(clist, &info);

   /* Position the normal working hour rows in the middle */
/* this doesn't work !!!!
   gtk_clist_thaw(GTK_CLIST(clist));
   gtk_clist_moveto(GTK_CLIST(clist), (gint)(20), 0, 0.5, 0);
*/
}
#endif


/******** Internal function definitions ********/

static void make_date_display(GtkWidget *tbox)
{
   static GtkWidget *label;
   gchar buf[40];

   if(label == NULL)
   {
      /* Get current system time */
      info.tref = time(&info.tref);
      set_current_time(&info);

      /* Set the row index to the time */
      info.row = info.hour - HOUR_BEGIN;

      /* Round down time to the hour */
      info.tref -= info.minute * 60;
      info.minute = 0;

      /* Initialise summary and details strings */
      info.summary = g_strdup("");
      info.details = g_strdup("");

      /* Create a text label for showing the current date */
      strftime(buf, sizeof buf, "%a %e %b %G", localtime(&info.tref));
      label = gtk_label_new(buf);
      gtk_box_pack_start(GTK_BOX(tbox), label, FALSE, FALSE, 5);
      gtk_widget_show(label);
   }
   else
   {
      /* Update the text label for the date */
      strftime(buf, sizeof buf, "%a %e %b %G", localtime(&info.tref));
      gtk_label_set_text(GTK_LABEL(label), buf);
   }

   /* setup data in the schedule list */
   if(clist != NULL)
      read_task_for_day(clist, &info);
}



static void change_day(GtkWidget *widget, gint mode)
{
   gint offset;

   switch(mode)
   {
      case YESTERDAY: offset = -1; break;
      case TODAY    : offset =  0; break;
      case TOMORROW : offset = +1; break;
      default: return;
   }

//   close_summary_entry();
//   clear_entry_buffer();

   info.tref += offset * 24 * 3600;

   set_current_time(&info);

   make_date_display(NULL);
}



static void init_list_items()
{
   GdkColor darkgrey, lightgrey;
   gchar *strval[2], buf0[10], buf1[60];
   int i, j;

   get_grey_colors(&darkgrey, &lightgrey);

   strval[0] = buf0;
   strval[1] = buf1;

   strcpy(strval[1], "");

   /* create rows for 24 hours only, really should have done better */
   for(i = HOUR_BEGIN, j = 0; i <= HOUR_END; i++, j++)
   {
      /* Show the time like hh:mm */
      sprintf(strval[0], "%2d:00", i);
   
      /* Add the data to the clist */
      gtk_clist_append(GTK_CLIST(clist), strval);

      if(i < WORK_BEGIN || (i >= LUNCH_BEGIN && i < LUNCH_END) || i > WORK_END)
         gtk_clist_set_background(GTK_CLIST(clist), j, &darkgrey);
      else
         gtk_clist_set_background(GTK_CLIST(clist), j, &lightgrey);
   }
}


static void edit_details(GtkWidget *widget, gpointer gdata)
{
//   close_summary_entry();

   make_event_details();
}



static void select_date(GtkWidget *widget, gpointer gdata)
{
//   close_summary_entry();

   make_calendar();
}



