#include "scheduler.h"


#define TOPLEVEL_WD             320
#define TOPLEVEL_HT             440


#define SPACING         	4



/******** Global variable declarations ********/
static GtkWidget *toplevel;
static GtkWidget *layout;
static GtkWidget *vbox;
GdkColormap *defcmap;
GtkStyle *defstyle, *cstyle;
GdkColor black, white;
GdkColor grey1 = {0, 0xcfff, 0xcfff, 0xcfff};
GdkColor grey2 = {0, 0xefff, 0xefff, 0xefff};
gint defsize;


/******** External function declarations ********/

void make_event_date(GtkWidget *vbox);
void make_event_list(GtkWidget *vbox);
void make_event_details();
void make_calendar();
void init_task_queue();
void close_summary_entry();
void close_scheduler(GtkWidget *widget, gpointer gdata);
void scheduler_resize_details( gint width, gint height );
void scheduler_resize_calendar( gint width, gint height );



/******** Internal function declarations ********/

static void get_system_defaults();
static void make_main_panel(GtkWidget *vbox);
gboolean scheduler_configure_event_callback( GtkWidget *widget, GdkEventConfigure *e, gpointer data );



/******** External function definitions ********/

void get_grey_colors(GdkColor *darkgrey, GdkColor *lightgrey)
{
   *darkgrey = grey1;
   *lightgrey = grey2;
}



GtkWidget *get_toplevel_widget()
{
   return toplevel;
}



GtkWidget *get_layout_widget()
{
   /* return the main container */
   return layout;
}



void show_main_window(gint flg)
{
   if(flg)
      gtk_widget_map(vbox);
   else
      gtk_widget_unmap(vbox);
}



void set_gc_black_white(GdkGC *fg, GdkGC *bg)
{
   gdk_gc_set_foreground(fg, &black);
   gdk_gc_set_background(bg, &white);
}



void close_scheduler(GtkWidget *widget, gpointer gdata)
{
   /* quit application here */
   gtk_main_quit();
}


#ifdef BUSYBOX
int scheduler_main(int argc, char **argv)
#else
int main(int argc, char **argv)
#endif
{
   /* Initialise the toolkit */
   gtk_set_locale();
   gtk_init(&argc, &argv);
   gtk_rc_init();

   /* get GC, fonts and colours etc */
   get_system_defaults();

   toplevel = gtk_window_new(GTK_WINDOW_TOPLEVEL);
   gtk_container_border_width(GTK_CONTAINER(toplevel), 0);
   gtk_signal_connect(GTK_OBJECT(toplevel), "destroy", \
                      GTK_SIGNAL_FUNC(close_scheduler), NULL);
   gtk_signal_connect(GTK_OBJECT(toplevel), "configure_event", \
                      GTK_SIGNAL_FUNC(scheduler_configure_event_callback), NULL);

   gtk_window_set_title(GTK_WINDOW(toplevel), TOPLEVEL_NM);
   gtk_widget_set_usize(GTK_WIDGET(toplevel), TOPLEVEL_WD, TOPLEVEL_HT);

   /* make the application visible */
   gtk_widget_show(toplevel);

   /* Create a fixed box an the main container */
   layout = gtk_fixed_new();
   gtk_widget_show(layout);
   gtk_container_add(GTK_CONTAINER(toplevel), layout);

   /* Create a vertical box for layout control */
   vbox = gtk_vbox_new(FALSE, ALL_SPACE);
   gtk_widget_show(vbox);
   gtk_fixed_put(GTK_FIXED(layout), vbox, 0, 0);

   /* create date display area and change day buttons */
   make_event_date(vbox);

   /* create clist for time and summary display */
   make_event_list(vbox);

   /* create user action area */
   make_main_panel(vbox);

   /* start the scheduler timer */
   init_task_queue();

   /* start the event main loop */
   gtk_main();

   return 0;
}



/******** Internal function definitions ********/

static void get_system_defaults()
{
   /* Get default colormap and depth */
   defcmap = gdk_colormap_get_system();
   defsize = gdk_colormap_get_system_size();

   /* Get black and white color values */
   gdk_color_black(defcmap, &black);
   gdk_color_white(defcmap, &white);

   /* Allocate colors */
   gdk_color_alloc(defcmap, &grey1);
   gdk_color_alloc(defcmap, &grey2);

   /* Get default style and make a copy */
   defstyle = gtk_widget_get_default_style();
   cstyle = gtk_style_copy(defstyle);
}



static void make_main_panel(GtkWidget *vbox)
{
   GtkWidget *cbox;

   // Button creation is now moved to list.c.

   cbox = gtk_hbox_new(FALSE, ALL_SPACE);
   gtk_widget_show(cbox);
   gtk_box_pack_start(GTK_BOX(vbox), cbox, FALSE, FALSE, THIRD_SPACE);
}




gboolean
scheduler_configure_event_callback( GtkWidget *widget, GdkEventConfigure *e, gpointer data ) {
   gtk_widget_set_usize( vbox, e->width, e->height );
   scheduler_resize_details( e->width, e->height );
   scheduler_resize_calendar( e->width, e->height );
   return FALSE;
}
