#!/bin/sh

#
# NanoGTK on PC X-Windows setup
#

check_dir ()
{
   if [ ! -d $1 ]; then
      echo
      echo "Could not find directory $1"
      echo
      exit 0
   fi
   echo "Found $1"
}

link_file ()
{
   if [ -h $2 ]; then
      rm -f $2
   fi

   if [ -f $1 -o -d $1 ]; then
      echo "ln -s $1 $2"
      ln -s $1 $2
   fi
}

copy_file ()
{
   if [ -f $1 ]; then
      echo "cp -pf $1 $2"
      cp -pf $1 $2
   else
      echo
      echo "Could not find $1"
      echo
      exit 0
   fi
}

exec_cmd ()
{
   echo "$*"
   $*
}


#
# save current working directory
#
THISDIR=`pwd`


#
# locate NanoGTK installation directory
#
if [ "$NANOGTK_HOME" = "" ]; then
   INSTDIR=`dirname \`pwd\``
else
   INSTDIR=$NANOGTK_HOME
fi

check_dir $INSTDIR


#
# set testing directory
#
TESTDIR=$INSTDIR/Testbench

check_dir $TESTDIR


echo
echo "<<< Setup NanoGTK Emulation environment for X11 >>>"
exec_cmd cd $TESTDIR


#
# check installed library and application directories
#
check_dir $INSTDIR/glib-1.2.7
check_dir $INSTDIR/gtk+-1.2.7
check_dir $INSTDIR/microwin
check_dir $INSTDIR/zlib
check_dir $INSTDIR/Application


#
# create various directories
#
exec_cmd cd $TESTDIR

for d in programs
do
   if [ ! -d $d ]; then
      exec_cmd mkdir $d
   else
      echo "directory $d already exists"
   fi
done


MCWDIR=$INSTDIR/microwin
APPDIR=../Application
BINDIR=bin-x86


#
# copy nano-X executables
#
copy_file $MCWDIR/$BINDIR/nanowm .
copy_file $MCWDIR/$BINDIR/nano-X .


#
# create symbolic link to config, icon and executable files
#
pdirs=`ls $APPDIR`


#
# Handle other applications
#
for appd in $pdirs
do
   d=$APPDIR/$appd

   # do not handle desktop here
   if [ -d $d/$BINDIR ]; then
         # link executables
         files=`ls $d/$BINDIR`

         if [ -n "$files" ]; then
            for fn in $files
            do
               if [ -x $d/$BINDIR/$fn ]; then
                  exec_cmd cd programs
                  link_file ../$d/$BINDIR/$fn $fn
                  exec_cmd cd ..
               fi
            done
         fi
   fi
done


cd $THISDIR

