/* GDK - The GIMP Drawing Kit
 * Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * Modified by the GTK+ Team and others 1997-1999.  See the AUTHORS
 * file for a list of people on the GTK+ Team.  See the ChangeLog
 * files for a list of changes.  These files are distributed with
 * GTK+ at ftp://ftp.gtk.org/pub/gtk/. 
 */

#include <string.h>
#include "gdk.h"
#include "gdkprivate.h"

#define PRINTFILE(s) g_message("gdkgc.c:%s()", s)
GdkGC*
gdk_gc_new (GdkWindow *window)
{
 return gdk_gc_new_with_values (window, NULL, 0);
}

GdkGC*
gdk_gc_new_with_values (GdkWindow	*window,
			GdkGCValues	*values,
			GdkGCValuesMask	 values_mask)
{
#ifdef UNUSED_WINDOW
  GdkWindowPrivate *window_private;
#endif
  GdkGC *gc;
  GdkGCPrivate *private;
  GR_GC_INFO nvalues;
  GR_LINE_ATTRIB_MASK gr_line_attrib_mask = 0;

#ifdef UNUSED_WINDOW
Is this window arg used somewhere ? Remove it.
  g_return_val_if_fail (window != NULL, NULL);

  window_private = (GdkWindowPrivate*) window;
  if (window_private->destroyed)
    return NULL;
#endif

  private = g_new (GdkGCPrivate, 1);
  gc = (GdkGC*) private;

  private->ref_count = 1;
  nvalues.gcid = GrNewGC(); 

  if (values_mask & GDK_GC_CLIP_X_ORIGIN)
    private->clip_x_origin = values->clip_x_origin;
  if (values_mask & GDK_GC_CLIP_Y_ORIGIN)
    private->clip_y_origin = values->clip_y_origin;
  if (values_mask & GDK_GC_TS_X_ORIGIN)
    private->ts_x_origin = values->ts_x_origin;
  if (values_mask & GDK_GC_TS_Y_ORIGIN)
    private->ts_y_origin = values->ts_y_origin;

  if (values_mask & GDK_GC_FOREGROUND)
    {
      	nvalues.foreground = values->foreground.pixel;
	GrSetGCForeground(nvalues.gcid, nvalues.foreground);
    }
  if (values_mask & GDK_GC_BACKGROUND)
    {
      	nvalues.background = values->background.pixel;
    	GrSetGCBackground(nvalues.gcid, nvalues.background);
    }
  if ((values_mask & GDK_GC_FONT) && (values->font->type == GDK_FONT_FONT))
    {
      nvalues.font = ((GdkFontPrivate*) (values->font))->fid;
	GrSetGCFont(nvalues.gcid, nvalues.font);
    }

    if (values_mask & GDK_GC_FUNCTION)
    {
      switch (values->function)
        {
	case GDK_INVERT:
                nvalues.mode= GR_MODE_INVERT;
                break;
        case GDK_XOR :
                nvalues.mode= GR_MODE_XOR;
                break;
        case GDK_OR :
                nvalues.mode= GR_MODE_OR;
                break;
        case GDK_AND :
                nvalues.mode= GR_MODE_AND;
                break;
        default :
                nvalues.mode = GR_MODE_SET;
        }
        GrSetGCMode(nvalues.gcid, nvalues.mode);
    }

	/*
	 Disable this for now, it conflicts with FUNCTION
	*/
  if (values_mask & GDK_GC_LINE_STYLE)
    {
      switch (values->line_style)
	{
	case GDK_LINE_SOLID:
	/*  nvalues.mode= GR_MODE_SET; */
	  	nvalues.line_style = GR_LINE_STYLE_SOLID;
	  	break;
	case GDK_LINE_ON_OFF_DASH:
	/*  nvalues.mode= GR_MODE_XOR; */
		nvalues.line_style = GR_LINE_STYLE_DASH;
	  	break;
	case GDK_LINE_DOUBLE_DASH:
	/*  nvalues.mode= GR_MODE_OR; */
		nvalues.line_style = GR_LINE_STYLE_DASH;
	  	break;
	}
	/* GrSetGCMode(nvalues.gcid, nvalues.mode); */
	gr_line_attrib_mask |= GR_LINE_ATTRIB_STYLE_SET;
    }
    if (values_mask & GDK_GC_LINE_WIDTH)
    {
        nvalues.line_width = values->line_width;
	gr_line_attrib_mask |= GR_LINE_ATTRIB_WIDTH_SET;
    }
    if (gr_line_attrib_mask != 0) 
	GrSetGCLineAttrib(nvalues.gcid, gr_line_attrib_mask, 
			nvalues.line_width, nvalues.line_style);
    if (values_mask & GDK_GC_SUBWINDOW)
    {
        nvalues.subwindowmode = values->subwindow_mode;
	GrSetGCSubWindowMode(nvalues.gcid, nvalues.subwindowmode);
    }

  private->xgc= nvalues.gcid;
#ifdef UNUSED_WINDOW
 private->xwindow= window_private->xwindow;
#else
 private->xwindow= 0;
#endif
//printf("New GC %d private %d and %d, %d\n", nvalues.gcid, private->xwindow, window_private->width, window_private->height);
  return gc;
}

void
gdk_gc_destroy (GdkGC *gc)
{
  gdk_gc_unref (gc);
}

GdkGC *
gdk_gc_ref (GdkGC *gc)
{
  GdkGCPrivate *private = (GdkGCPrivate*) gc;

  g_return_val_if_fail (gc != NULL, NULL);
  private->ref_count += 1;

  return gc;
}

void
gdk_gc_unref (GdkGC *gc)
{
  GdkGCPrivate *private = (GdkGCPrivate*) gc;
 
  g_return_if_fail (gc != NULL);
  g_return_if_fail (private->ref_count > 0);
 
  if (private->ref_count > 1)
    private->ref_count -= 1;
  else
    {
	GrDestroyGC(GDK_GC_XGC(gc));
      memset (gc, 0, sizeof (GdkGCPrivate)); // why is this needed ??
      g_free (gc);
    }

}

void
gdk_gc_get_values (GdkGC       *gc,
		   GdkGCValues *values)
{
  GR_GC_INFO gcinfo;
  GR_FONT_INFO fi;
  GdkFontPrivate private;

  g_return_if_fail (gc != NULL);
  g_return_if_fail (values != NULL);
  GrGetGCInfo(GDK_GC_XGC(gc), &gcinfo);
  if(gcinfo.gcid)
  {
	values->foreground.pixel = gcinfo.foreground;
	values->background.pixel = gcinfo.background;
	values->font = gdk_font_lookup(gcinfo.font);

	if(!values->font)
	{
		g_warning("Unable to load font from cache - get from Server!");
		GrGetFontInfo(gcinfo.font, &fi);
		private.xfont = &fi;
        	private.fid = gcinfo.font;
        	private.ref_count = 1;
        	private.names = NULL;

        	values->font = (GdkFont*) &private;
        	values->font->type = GDK_FONT_FONT;
        	values->font->ascent =  fi.baseline;
        	values->font->descent = fi.height-fi.baseline;
	}
	
	switch(gcinfo.mode)
	{
        	case GDK_XOR :
                	values->function = GR_MODE_XOR;
                	break;
        	case GDK_OR :
                	values->function = GR_MODE_OR;
                	break;
        	case GDK_AND :
                	values->function = GR_MODE_AND;
                	break;
        	default :
                	values->function = GR_MODE_SET;
			break;
	}
	values->subwindow_mode = gcinfo.subwindowmode;
#ifdef LINE_ATTRIB_SUPPORT
	values->line_width = gcinfo.line_width;
	switch (gcinfo.line_style) {
	case GR_LINE_STYLE_SOLID:
			values->line_style = GDK_LINE_SOLID;
			break;
	case GR_LINE_STYLE_DASH:
			values->line_style = GDK_LINE_ON_OFF_DASH;
			break;
	}
#endif
  }
  else
      memset (values, 0, sizeof (GdkGCValues));
}

void
gdk_gc_set_foreground (GdkGC	*gc,
		       GdkColor *color)
{
  g_return_if_fail (gc != NULL);
  g_return_if_fail (color != NULL);
  GrSetGCForeground(GDK_GC_XGC(gc),color->pixel);
}

void
gdk_gc_set_background (GdkGC	*gc,
		       GdkColor *color)
{
  g_return_if_fail (gc != NULL);
  g_return_if_fail (color != NULL);
  GrSetGCBackground(GDK_GC_XGC(gc),color->pixel);
}


void
gdk_gc_set_font (GdkGC	 *gc,
		 GdkFont *font)
{
  GdkFontPrivate *font_private;

  g_return_if_fail (gc != NULL);
  g_return_if_fail (font != NULL);

  if (font->type == GDK_FONT_FONT)
    {
      font_private = (GdkFontPrivate*) font;
	GrSetGCFont(GDK_GC_XGC(gc), font_private->fid);
    }

}

void
gdk_gc_set_function (GdkGC	 *gc,
		     GdkFunction  function)
{
	int mode;
	g_return_if_fail (gc != NULL);

	switch (function)
        {
        case GDK_INVERT :
                mode= GR_MODE_INVERT;
                break;
        case GDK_XOR :
                mode= GR_MODE_XOR;
                break;
        case GDK_OR :
                mode= GR_MODE_OR;
                break;
        case GDK_AND :
                mode= GR_MODE_AND;
                break;
        default :
                mode = GR_MODE_SET;
        }
        GrSetGCMode(GDK_GC_XGC(gc), mode);

}

void
gdk_gc_set_fill (GdkGC	 *gc,
		 GdkFill  fill)
{
  GdkGCValues values;

  g_return_if_fail (gc != NULL);

  values.fill = fill;
/*FIXME : nanox no  GDK_GC_FILL*/
}

void
gdk_gc_set_tile (GdkGC	   *gc,
		 GdkPixmap *tile)
{
  GdkGCValues values;

  g_return_if_fail (gc != NULL);

  values.tile = tile;
  /*FIXME: nanox no GDK_GC_TILE */

}

void
gdk_gc_set_stipple (GdkGC     *gc,
		    GdkPixmap *stipple)
{
  GdkGCValues values;

  g_return_if_fail (gc != NULL);

  values.stipple= stipple; 
  /*FIXME: nanox no GDK_GC_TILE */

}

void
gdk_gc_set_ts_origin (GdkGC *gc,
		      gint   x,
		      gint   y)
{
 GdkGCPrivate *private;
  g_return_if_fail (gc != NULL);

  private = (GdkGCPrivate*)gc;
  private->ts_x_origin = x;
  private->ts_y_origin = y;
}

void
gdk_gc_set_clip_origin (GdkGC *gc,
			gint   x,
			gint   y)
{
  GdkGCPrivate *private;

  g_return_if_fail (gc != NULL);

  private = (GdkGCPrivate*)gc;


   private->clip_x_origin = x;
   private->clip_y_origin = y;
}

void
gdk_gc_set_clip_mask (GdkGC	*gc,
		      GdkBitmap *mask)
{
	g_return_if_fail (gc != NULL);

	/* FIXME: Didn't take care about the clipmask, always set the clipmask to
	   zero. */
	if(!mask)
		GrSetGCRegion(GDK_GC_XGC(gc), 0);
/*	else
	{
		GR_WINDOW_INFO maskinfo;
	 	GR_RECT cliprect;

	        if(GDK_DRAWABLE_DESTROYED(mask)) return;

		GrGetWindowInfo(GDK_DRAWABLE_XID(mask), &maskinfo);
      		cliprect.x = 0;
	      	cliprect.y = 0;
      		cliprect.width = maskinfo.width;
	      	cliprect.height = maskinfo.height;
		GrSetClipRectangles(GDK_GC_XGC(gc), 0, 0, 1, &cliprect);
	}*/

}


void
gdk_gc_set_clip_rectangle (GdkGC	*gc,
			   GdkRectangle *rectangle)
{
  if(rectangle)
  {
	GR_RECT cliprect;
      	cliprect.x = rectangle->x ;
      	cliprect.y = rectangle->y;
      	cliprect.width = rectangle->width;
      	cliprect.height = rectangle->height;
	GrSetClipRectangles(GDK_GC_XGC(gc), 0, 0, 1, &cliprect);
    }

  else
	GrSetGCRegion(GDK_GC_XGC(gc), 0);
} 

void
gdk_gc_set_clip_region (GdkGC		 *gc,
			GdkRegion	 *region)
{
	g_return_if_fail (gc != NULL);
	if (region)
	{
		GdkRegionPrivate *region_private = (GdkRegionPrivate*) region;
#if 1
		GrSetGCRegion_X(GDK_GC_XGC(gc), region_private->xregion);
#else
		Region		xregion = region_private->xregion;
		GR_RECT		rect[xregion->numRects];
		int		i;
		
		for (i=0;i<xregion->numRects;i++)
		{
			rect[i].x = xregion->rects[i].x1;
			rect[i].y = xregion->rects[i].y1;
			rect[i].width = xregion->rects[i].x2 - xregion->rects[i].x1;
			rect[i].height = xregion->rects[i].y2 - xregion->rects[i].y1;
		}
		//GrSetGCRegion(GDK_GC_XGC(gc), region_private->xregion);
		GrSetClipRectangles(GDK_GC_XGC(gc), 0, 0, region_private->xregion->numRects, rect);
#endif
	}
	else
		GrSetGCRegion(GDK_GC_XGC(gc), 0);
}

void
gdk_gc_set_subwindow (GdkGC	       *gc,
		      GdkSubwindowMode	mode)
{
  g_return_if_fail (gc != NULL);

  GrSetGCSubWindowMode(GDK_GC_XGC(gc), mode);
}

void
gdk_gc_set_exposures (GdkGC     *gc,
		      gboolean   exposures)
{
	GdkGCValues values;

  	g_return_if_fail (gc != NULL);

  	values.graphics_exposures = exposures;
	PRINTFILE("__UNSUPPORTED__ : gdk_gc_set_exposures");
	/*FIXME : nanox have no GDK_GC_EXPOSURES */
}

void
gdk_gc_set_line_attributes (GdkGC	*gc,
			    gint	 line_width,
			    GdkLineStyle line_style,
			    GdkCapStyle	 cap_style,
			    GdkJoinStyle join_style)
{
	GR_SIZE gr_line_width = line_width;
	GR_LINE_STYLE gr_line_style;
	GR_LINE_ATTRIB_MASK value_mask = GR_LINE_ATTRIB_WIDTH_SET | GR_LINE_ATTRIB_STYLE_SET;

	g_return_if_fail (gc != NULL);
	switch (line_style) {
	case GDK_LINE_SOLID: 
			gr_line_style = GR_LINE_STYLE_SOLID;
			break;
	case GDK_LINE_ON_OFF_DASH: 
			gr_line_style = GR_LINE_STYLE_DASH;
			break;
	case GDK_LINE_DOUBLE_DASH: 
			gr_line_style = GR_LINE_STYLE_DASH;
	                PRINTFILE("__UNSUPPORTED__ : gdk_gc_set_line_attributes : GDK_LINE_DOUBLE_DASH");
			break;
	}
	GrSetGCLineAttrib(GDK_GC_XGC(gc), value_mask, gr_line_width, gr_line_style);
	
	PRINTFILE("__UNSUPPORTED__ : gdk_gc_set_line_attributes");
}

void
gdk_gc_set_dashes (GdkGC      *gc,
		   gint	       dash_offset,
		   gint8       dash_list[],
		   gint        n)
{
	GR_COUNT n_dashes;
	GR_DASH_SIZE *gr_dash_list;
	gint	actual_offset = dash_offset % n;
	gint 	i;

	/* take care of offset now */
	if (actual_offset < 0)
		return;
	n_dashes = n - dash_offset;
	gr_dash_list = g_new(GR_DASH_SIZE, n_dashes);
	for (i = 0; i < n_dashes; i++)
		gr_dash_list[i] = dash_list[i];
	GrSetGCLineDashSizes(GDK_GC_XGC(gc), n_dashes, gr_dash_list);
	if (gr_dash_list != NULL)
		g_free(gr_dash_list);
}

void
gdk_gc_copy (GdkGC *dst_gc, GdkGC *src_gc)
{
	GrDestroyGC(GDK_GC_XGC(dst_gc));
	GDK_GC_XGC(dst_gc) = GrCopyGC(GDK_GC_XGC(src_gc));
}
