#ifndef client_gtkC_c
#define client_gtkC_c

typedef struct {
        int x1, y1, x2, y2;
} BOX;

typedef struct {
        long size;
        long numRects;
        BOX *rects;
        BOX extents;
} Region;


#ifdef LINE_ATTRIB_SUPPORT
/*
 * Set the line width in a graphics context.
 */
void
GrSetGCLineAttrib(GR_GC_ID gc, GR_LINE_ATTRIB_MASK mask, GR_SIZE line_width, GR_LINE_STYLE line_style)
{
        nxSetGCLineAttribReq *req;
        if (line_width < 0)
           return;
        if (line_style < 0)
           return;
        req = AllocReq(SetGCLineAttrib);
        req->gcid = gc;
        req->mask = mask;
        req->line_width = line_width;
        req->line_style = line_style;
}

void
GrSetGCLineDashSizes(GR_GC_ID gc, GR_COUNT n, GR_DASH_SIZE *dashes)
{
        nxSetGCLineDashSizesReq *req;
        long            size;
        size = (long)n* sizeof(GR_DASH_SIZE);
        req = AllocReqExtra(SetGCLineDashSizes, size);
        req->gcid = gc;
        // Break into MAXREQUESTSZ ??
        memcpy(GetReqData(req), dashes, size);

}
#endif

/*
 * Set the subwindow mode in a graphics context.
 */
void
GrSetGCSubWindowMode(GR_GC_ID gc, int subwindowmode)
{
        nxSetGCSubWindowModeReq *req;

        if(subwindowmode > GR_INCLUDE_INFERIORS)
                return;

        req = AllocReq(SetGCSubWindowMode);
        req->gcid = gc;
        req->subwindowmode = subwindowmode;
}

void
GrSetSelectionOwner_aux(GR_WINDOW_ID owner, GR_ATOM_ID atom, unsigned long time)
{
        nxSetSelectionOwnerReq *req;

        req = AllocReq(SetSelectionOwner);
        req->owner= owner;
        req->atom = atom;
        req->time = time ;
}

GR_WINDOW_ID
GrGetSelectionOwner_aux(GR_ATOM_ID atom)
{
        GR_WINDOW_ID wid;
        nxGetSelectionOwner_auxReq *req;

        req =  AllocReq(GetSelectionOwner_aux);
        req->atom = atom;
        if(GrTypedReadBlock(&wid, sizeof(GR_WINDOW_ID),GrNumGetSelectionOwner) == -1)
                return 0;
        return wid;
}

int
GrGetFontList_aux(const char *name, char **list[])
{
        nxGetFontList_auxReq        *req;
        int                     size = strlen(name);
        int                     retval;
        char                    **p_l;
        int                     len;
        int                     i;

        req = AllocReqExtra(GetFontList_aux, size);
        memcpy(GetReqData(req), name, size);
        GrTypedReadBlock(&retval, sizeof(retval),GrNumGetFontList);

        if (retval<=0) return retval;

        p_l = *list = (char **)malloc((retval+1)*sizeof(char *));
        p_l[retval] = NULL;

        for (i=0;i<retval;i++)
        {
                if (GrReadBlock(&len,sizeof(int) ) == -1)
                        return 0;
                p_l[i] = (char *)malloc(len * sizeof(char) + 1);
                if (GrReadBlock(p_l[i],len * sizeof(char)) == -1)
                        return 0;
                p_l[i][len] = '\0';
        }

        return retval;
}

void
GrMoveResizeWindow(GR_WINDOW_ID wid, GR_COORD x, GR_COORD y, GR_SIZE width, GR_SIZE height)
{
        nxMoveResizeWindowReq *req;

        req = AllocReq(MoveResizeWindow);
        req->windowid = wid;
        req->x = x;
        req->y = y;
        req->width =width ;
        req->height =height ;

}

void
GrAreaSrc(GR_DRAW_ID id, GR_GC_ID gc, GR_COORD x, GR_COORD y, GR_SIZE width,
        GR_SIZE height, GR_COORD src_x, GR_COORD src_y, GR_SIZE src_width ,
        void *pixels, int pixtype, short alpha_value)
{
        nxAreaSrcReq    *req;
        long            size;
        long            chunk_y;
        int             pixsize;
        char            *small_buffer;
        int             h;
        long            src_rowstride = 0;
        long            rowstride = 0;

        /* Calculate size of packed pixels*/
        switch(pixtype) {
        case MWPF_RGB:
                pixsize = sizeof(MWCOLORVAL);
                break;
        case MWPF_PIXELVAL:
                pixsize = sizeof(MWPIXELVAL);
                break;
        case MWPF_PALETTE:
        case MWPF_TRUECOLOR332:
                pixsize = sizeof(unsigned char);
                break;
        case MWPF_TRUECOLOR0888:
                pixsize = sizeof(unsigned long);
                break;
        case MWPF_TRUECOLOR888:
                pixsize = 3;
                break;
        case MWPF_TRUECOLOR565:
                pixsize = sizeof(unsigned short);
                break;
        case MWPF_TRUECOLOR444:
        case MWPF_TRUECOLOR444_PACKED:
                pixsize = sizeof(unsigned short);
                break;
        case MWPF_GRAY_PACKED_4:
        case MWPF_GRAY_PACKED_4_ALPHA:
                pixsize = sizeof(unsigned char);
                break;
        default:
                return;                                                          
        }

        if (pixtype == MWPF_GRAY_PACKED_4)
        {
                if (width&1)
#ifdef PACKED_LINES
                        rowstride = width >> 1;
#else
                        rowstride = (width >> 1) +1 ;
#endif
                else
                        rowstride = width >> 1;
                if (src_width&1)
#ifdef PACKED_LINES
                        src_rowstride = src_width >> 1;
#else
                        src_rowstride = (src_width >> 1) + 1;
#endif
                else
                        src_rowstride = src_width >> 1;
                pixels += src_rowstride * src_y + (src_x >> 1); //keep the first half byte.
        }
        else
        if (pixtype == MWPF_TRUECOLOR444_PACKED)
        {
                //rowstride=((width+1)*3)/2;
                //src_rowstride=((src_width+1)*3)/2;
                rowstride=(width*3)/2 + (width&1);
                src_rowstride=(src_width*3)/2 + (src_width&1);
                pixels += src_rowstride * src_y +  (src_x>>1)*3 +(src_x&1);
                if (src_x&1) {
                        pixtype|=OFFSET_BYTES;
                        if (1-(width&1))
                                rowstride++;
                }
        }
        else
        {
                rowstride=width * pixsize;
                src_rowstride=src_width * pixsize;
                pixels += src_rowstride * src_y + src_x*pixsize;
        }
        /* Break request into MAXREQUESTSZ size packets*/
        while(height > 0) {
                chunk_y = (rowstride == 0) ? height: (MAXREQUESTSZ - sizeof(nxAreaSrcReq)) / rowstride;
                if(chunk_y > height)
                        chunk_y = height;
                size = chunk_y * rowstride;                                      
                req = AllocReqExtra(AreaSrc, size);
                req->drawid = id;
                req->gcid = gc;
                req->x = x;
                req->y = y;
                req->width = width;
                req->height = chunk_y;
                req->pixtype = pixtype;
                req->alpha_value = alpha_value;
                small_buffer = GetReqData(req);
                h = 0;
                do
                {
#ifdef PACKED_LINES
                        char * dst;
                        if (h&1)
                        {
                                if (src_width&1)
                                        src_rowstride++;
                                if (width&1)
                                        rowstride++;
                        }
                        else if (h>0)
                        {
                                if (src_width&1)
                                        src_rowstride--;
                                if (width&1)
                                        rowstride--;
                        }
                        dst = small_buffer + h*rowstride;
                        memcpy(dst,pixels,rowstride);
#else
                        memcpy(&small_buffer[h*rowstride],pixels,rowstride);
#endif                                                                           
                        pixels +=src_rowstride;
                        h++;
                }
                while (h<chunk_y);
                y += chunk_y;
                height -= chunk_y;
        }
}                                                                                
void
GrSegments(GR_DRAW_ID id, GR_GC_ID gc, GR_COUNT nsegs, GR_SEGMENT *segs)
{
        nxSegmentsReq   *req;
        long                    size;

        if(nsegs == 0 || segs == NULL)
                return;

        size = (long)nsegs* sizeof(GR_SEGMENT);
        req = AllocReqExtra(Segments, size);
        req->drawid = id;
        req->gcid = gc;
        // Break into MAXREQUESTSZ ??
        memcpy(GetReqData(req), segs, size);
}

void
GrClearChildArea(GR_WINDOW_ID wid, GR_COORD x, GR_COORD y,
        GR_SIZE width, GR_SIZE height)
{
        nxClearChildAreaReq *req;
        req = AllocReq(ClearChildArea);
        req->wid = wid;
        req->x = x;
        req->y = y;
        req->width = width;
        req->height = height;
}

void
GrScrollWindow(GR_DRAW_ID id, GR_GC_ID gc, GR_COORD offx, GR_COORD offy)
{
        nxScrollWindowReq *req;

        req = AllocReq(ScrollWindow);
        req->wid = id;
        req->gcid = gc;
        req->offx = offx;
        req->offy = offy;
}

int
GrGetNextExposeWin(GR_EVENT *ep, GR_WINDOW_ID wid)
{
#ifdef USE_SOCKET
        EVENT_LIST * ptr;
        EVENT_LIST * prev = NULL;
        GR_EVENT * ev;

        GrFlush();

        ptr = evlist;

        while (ptr)
        {
                ev = &ptr->event;

                if ((ev->type == GR_EVENT_TYPE_EXPOSURE)
                    && ( (wid == -1)
                    || (ev->exposure.wid == wid) ))
                {
                        //memcpy(ep,ev,sizeof(GR_EVENT));
                        *ep = *ev;

                        if (ptr == evlist)
                        {
                                evlist=evlist->next;
//                                free(ptr->event);
                                free(ptr);
                        }
                        else
                        {
                                if (ptr->next!=NULL)
                                        prev->next = prev->next->next;
                                else
                                        prev->next = NULL;
//                                free(ptr->event);
                                free(ptr);
                        }
                        return 1;
                }
                prev = ptr;
                ptr = ptr->next;
        }
        ep->type = GR_EVENT_TYPE_NONE;

        return 0;
#else
        nxGetNextExposeWinReq *req;

        req = AllocReq(GetNextExposeWin);
        req->wid =wid;
        GrTypedReadBlock(ep, sizeof(*ep),GrNumGetNextExposeWin);
        return GrReadByte();
#endif
}

void
GrDeleteProp(GR_WINDOW_ID wid, GR_ATOM_ID prop_id)
{
        nxDeletePropReq *req;

        req = AllocReq(DeleteProp);
        req->wid = wid;
        req->prop_id = prop_id;
}

void
GrGetAtomName(GR_ATOM_ID atom, GR_CHAR **string)
{
        nxGetAtomNameReq *req;
        int len;
        char c;

        req = AllocReq(GetAtomName);
        req->atom = atom;
        GrTypedReadBlock(&len, sizeof(len),GrNumGetAtomName);
        if(!len) {
                *string = NULL;
                return;
        }
        if(!(*string = malloc(len))) {
                /* Oh dear, we're out of memory but still have to purge the
                   requested data (and throw it away) */
                while(len--) GrReadBlock(&c, 1);
        } else {
                GrReadBlock(*string, len);
        }
}

void
GrGetPointer(GR_WINDOW_ID wid, GR_WINDOW_ID *return_wid, GR_COORD *coordx, GR_COORD *coordy)
{
        nxGetPointerReq *req;

        req = AllocReq(GetPointer);
        req->windowid  = wid;
        GrTypedReadBlock(return_wid, sizeof(*return_wid),GrNumGetPointer) ;
        GrReadBlock(coordx, sizeof(*coordx));
        GrReadBlock(coordy, sizeof(*coordy));
}

void
GrSetGCRegion_X(GR_GC_ID gcid, Region *region)
{
        nxSetGCRegionXReq *req;
        long       size;
        int     nb_rect = region->numRects;
        int     chunk;
        MWRECT *rects = (MWRECT *)region->rects;

        /* BOX is casted to MWRECT. Beware if one of this struct is changed */

        /* Break request into MAXREQUESTSZ size packets*/
        while(nb_rect > 0) {
                chunk = (MAXREQUESTSZ - sizeof(nxSetGCRegionXReq)) / sizeof(MWRECT);
                if(chunk > nb_rect)
                        chunk = nb_rect;
                size = chunk * sizeof(MWRECT);
                req = AllocReqExtra(SetGCRegionX, size);
                req->gcid = gcid;
                req->size = region->size;
                req->numRects = region->numRects;
                req->left = region->extents.x1;
                req->top = region->extents.y1;
                req->right = region->extents.x2;
                req->bottom = region->extents.y2;
                memcpy(GetReqData(req), rects, size);
                nb_rect -= chunk;
                rects+=chunk;
        }
}

/*
 * Set the background of a window to the specified color. Redraw if specified.
 */
void
GrSetBackgroundColor(GR_WINDOW_ID wid, GR_COLOR colour)
{
        nxSetBackgroundColorReq *req;

        req = AllocReq(SetBackgroundColor);
        req->windowid = wid;
        req->color = colour;
}

void
GrInjectSelectionEvent(GR_EVENT_SELECTION *event)
{
        nxInjectEventReq *req;

        req = AllocReq(InjectEvent);
        req->event_type = event->type;
        req->event.selection.owner = event->owner;
        req->event.selection.requestor = event->requestor;
        req->event.selection.selection = event->selection;
        req->event.selection.target = event->target;
        req->event.selection.property = event->property;
        req->event.selection.time = event->time;

        GrFlush();
}

void
GrLines(GR_DRAW_ID id, GR_GC_ID gc, GR_COUNT npoints, GR_POINT *points)
{
        nxLinesReq   *req;
        long                    size;

        if(npoints == 0 || points == NULL)
                return;

        size = (long)npoints* sizeof(GR_POINT);
        req = AllocReqExtra(Lines, size);
        req->drawid = id;
        req->gcid = gc;
        // Break into MAXREQUESTSZ ??
        memcpy(GetReqData(req), points, size);
}

void
GrChangeProp(GR_WINDOW_ID wid, GR_ATOM_ID prop_id, GR_ATOM_ID type, GR_SIZE format,
        GR_CHAR *data, GR_SIZE nitems)
{
        nxChangePropReq *req;

        if(nitems > MAXREQUESTSZ)
        {
                /* FIXME: this should be handle, but for now,
                        should let the program return gracefully */
                printf("Nano-X Warning: Data length is greater than
                        MAXREQUESTSZ\n");
                return;
        }
        req = AllocReqExtra(ChangeProp, nitems);
        req->wid = wid;
        req->prop_id = prop_id;
        req->type = type;
        req->format = format;
        req->nitems = nitems;

        memcpy((GR_CHAR*)GetReqData(req), data, nitems);
}

void
GrConvertSelection(GR_ATOM_ID atom, GR_ATOM_ID target, GR_ATOM_ID prop,
        GR_WINDOW_ID requestor, unsigned long time)
{
        nxConvertSelectionReq *req;

        req = AllocReq(ConvertSelection);
        req->atom = atom;
        req->requestor= requestor;
        req->target= target;
        req->prop = prop;
        req->time= time;
}

void
GrGetEncoding(int* encoding, GR_SIZE * width)
{
        nxGetEncodingReq        *req;
        req = AllocReq(GetEncoding);
        GrTypedReadBlock(encoding, sizeof(int), GrNumGetEncoding);
        GrReadBlock(width, sizeof(int));
}

void
GrGrabPointer(GR_WINDOW_ID wid, GR_EVENT_MASK mask) {
        // this allocates memory, initialise the request
        // and send a request to the server...
    nxGrabPointerReq *req;
    req = AllocReq(GrabPointer);
    req->wid  = wid;
    req->mask = mask;
}

GR_ATOM_ID
GrInternAtom(const GR_CHAR* atom_name, int only_if_exists)
{
        nxInternAtomReq *req;
        GR_ATOM_ID atom_found;
        int len;

        len = strlen(atom_name);

        if(!len)
                return -1;
        req = AllocReqExtra(InternAtom, len);
        req->only_if_exists = only_if_exists;
        memcpy(GetReqData(req), atom_name , len+1);
        if(GrTypedReadBlock(&atom_found, sizeof(GR_ATOM_ID),GrNumInternAtom) == -1)
                return -1;
        return atom_found;
}

void
GrSetEncoding(int encoding)
{
        nxSetEncodingReq        *req;
        req = AllocReq(SetEncoding);
        req->encoding = encoding;
}

void
GrSetClipRectangles(GR_GC_ID gcid, GR_COORD origin_x, GR_COORD origin_y, int n,
        GR_RECT* cliprect)
{
        nxSetClipRectanglesReq *req;
        long       size;
        int     nb_rect = n;
        int     chunk;

        /* Break request into MAXREQUESTSZ size packets*/
        while(nb_rect > 0) {
                chunk = (MAXREQUESTSZ - sizeof(nxSetClipRectanglesReq)) / sizeof(GR_RECT);
                if(chunk > nb_rect)
                        chunk = nb_rect;
                size = chunk * sizeof(GR_RECT);
                req = AllocReqExtra(SetClipRectangles, size);
                req->gcid = gcid;
                req->origin_x = origin_x;
                req->origin_y = origin_y;
                req->n = n;
                memcpy(GetReqData(req), cliprect, size);
                nb_rect -= chunk;
                cliprect+=chunk;
        }
}

void
GrGetWindowProp(GR_WINDOW_ID wid, GR_ATOM_ID prop_id, GR_ATOM_ID *ret_type,
        int *ret_format, unsigned long *ret_nitems, GR_CHAR **prop_return)
{
        nxGetWindowPropReq *req;
        GR_SIZE len;

        req = AllocReq(GetWindowProp);
        req->wid = wid;
        req->prop_id = prop_id;

        GrTypedReadBlock(ret_type, sizeof(*ret_type), GrNumGetWindowProp);
        GrReadBlock(ret_format, sizeof(*ret_format));
        GrReadBlock(ret_nitems, sizeof(*ret_nitems));
        if(*ret_nitems)
        {
                GrReadBlock(&len, sizeof(len));
                if(len)
                *prop_return = (GR_CHAR*)malloc(sizeof(GR_CHAR)* len);
                GrReadBlock(*prop_return, len);
        }
}

/*
 * [Simplified Chinese] - GBK
 *      First  byte range: 0x81-0xfe
 *      Second byte range: 0x40-0x7e, 80-fe
 */
MWBOOL
is_gbk_char (unsigned char c1, unsigned char c2)
{
        return (c1 >= 0x81 && c1 <= 0xfe
                && (  (c2 >= 0x40 && c2 <= 0x7e)
                   || (c2 >= 0x80 && c2 <= 0xfe))
               );
}

#ifdef HAVE_HWR_SUPPORT
void
GrExitDrawingMode()
{
        nxExitDrawingModeReq *req;

        req = AllocReq(ExitDrawingMode);

}
#endif

/*
 * Set the border of a window to the specified color.
 */
#if (0)
void
GrSetBorderColor(GR_WINDOW_ID wid, GR_COLOR colour)
{
        nxSetBorderColorReq *req;

        req = AllocReq(SetBorderColor);
        req->windowid = wid;
        req->color = colour;
}
#endif

/*
 * Specify a cursor for a window.
 * This cursor will only be used within that window, and by default
 * for its new children.  If the cursor is currently within this
 * window, it will be changed to the new one immediately.
 */
void
GrSetCursor_aux(GR_WINDOW_ID wid, GR_SIZE width, GR_SIZE height, GR_COORD hotx,
        GR_COORD hoty, GR_COLOR foreground, GR_COLOR background,
#ifdef NANOGTK
        int useparent,
#endif
        GR_BITMAP *fgbitmap, GR_BITMAP *bgbitmap)
{
        nxSetCursor_auxReq *req;
        int             bitmapsize;
        char *          data;

        bitmapsize = GR_BITMAP_SIZE(width, height) * sizeof(GR_BITMAP);
        req = AllocReqExtra(SetCursor_aux, bitmapsize*2);
        req->windowid = wid;
        req->width = width;
        req->height = height;
        req->hotx = hotx;
        req->hoty = hoty;
        req->fgcolor = foreground;
        req->bgcolor = background;
#ifdef NANOGTK
        req->useparent = useparent;
#endif
        data = GetReqData(req);
        memcpy(data, fgbitmap, bitmapsize);
        memcpy(data+bitmapsize, bgbitmap, bitmapsize);
}

void
GrSelectOnly(int *maxfd,void *rfdset)
{
/*
        fd_set *rfds = rfdset;

        FD_SET(nxSocket, rfds);
        if(nxSocket > *maxfd)
                *maxfd = nxSocket;
        if(regfd != -1) {
                FD_SET(regfd, rfds);
                if(regfd > *maxfd)
                        *maxfd = regfd;
        }
*/
}

#if (0)
void
GrGetEvent(GR_EVENT * ev)
{
        struct ev_list * ptr = event_queue;

        if (queue_length>0)
        {
                //memcpy(ev, event_queue->ev, sizeof(GR_EVENT));
                *ev = *event_queue->ev;
                free(event_queue->ev);
                event_queue = event_queue->next;
                free(ptr);
                queue_length--;
        }
        else
        {
                ev->type=GR_EVENT_TYPE_NONE;
        }
}
#endif

#endif // client_gtkC_c

