#ifdef NANOGTK
#if (1)
#include <signal.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "serv.h"

#define PROC_DIR        "/proc/"
#define MAX_CHAR        80
#define SLEEPING	'S'
#define RUNNING		'R'
#define STOPPED		'T'

static int wokeup_pid=0;
static GR_WINDOW *wokeup_win=NULL;

/* Function Prototype */
static void ltostr(long value, GR_CHAR** ret_str);
static GR_CHAR GetPidStatus(int pid);
void strreverse (GR_CHAR *string);


/*
 * GetPidStatus()
 * pid - id of a rpocess which status to be retrieved.
 * Get the process current status from the /proc/pid/status file.
 * R : (running)
 * T : (stopped)
 * S : (sleeping)
 */
static GR_CHAR 
GetPidStatus(int pid)
{
        FILE *fp;
        GR_CHAR filename[50]; /* should be "/proc/pid/status" */
        GR_CHAR *str_pid=NULL;
        GR_CHAR name[20];
        GR_CHAR val[40];
        GR_CHAR status=0;

        ltostr(pid, &str_pid);
        strcpy(filename, PROC_DIR);
        strcat(filename, str_pid);
        strcat(filename, "/status");

        if( (fp = fopen(filename, "r")))
        {
                while(!feof(fp))
                {
                	fscanf(fp,"%[^:\t] %*s %[^\n] " ,name, val);
			if(!strcmp(name, "State"))
			{
				status = val[0];
				break;
			}
                }
        }
        if(fp) fclose(fp);
        free(str_pid);
	return status;
}


/*
 * GsSendContSig()
 * wp - window which to be "wake up" if sleep.
 * This function is actually designed specific - for now.
 *  
 */
void
GsSendContSig(GR_WINDOW* wp)
{
        GR_WINDOW *toplevel;

	/* Reset everything */
        wokeup_win = NULL;
        wokeup_pid = 0;

        if(!wp)
                return;

        toplevel = GsFindTopLevelParent(wp);
        /* Top Level window is unmapped - pid suspended */
        if(toplevel->unmapcount > 0)
        {
                GR_EVENT_CLIENT *ecp;
		GR_CHAR	status;
                ecp = toplevel->eventclients;
                status = GetPidStatus(ecp->client->pid);

		/* If the client is currently active 
		 * don't send any CONT signal to it and 
		 * exit this function.
		 */
		if(status != STOPPED)
			return;

                kill(ecp->client->pid, SIGCONT);
                printf("Activate client %d\n", ecp->client->pid);
		/* Save the wokenup window and process , to be put
		 * to sleep later, if needed.
		 */
                wokeup_win = toplevel;
                wokeup_pid = ecp->client->pid;
        }
}

/* 
 * GsDeactivateWin()
 * This function will put an process into sleep. This process is
 * got woken up by the above functioni GsSendContSignal().
 * The criteria of determining the process is should be sleeping is
 * by checking the the top-level window is still unmapped - is
 * design specific. 
 * However, Desktop should not fall into this criteria.
 * Checking is done at the above function.
 */
void
GsDeactivateWin()
{
        if(wokeup_win)
        {
                /* still is suspended*/
                if(wokeup_win->unmapcount > 0)
                        kill(wokeup_pid, SIGSTOP);

                wokeup_win = NULL; 
        }
        wokeup_pid = 0;
}

/*
 * strreverse()
 * string - string to be reverse.
 * This function is actually borrow from the glib
 */
void
strreverse (GR_CHAR *string)
{


  if (string && *string)
    {
      register GR_CHAR *h, *t;

      h = string;
      t = string + strlen (string) - 1;

      while (h < t)
        {
          register GR_CHAR c;

          c = *h;
          *h = *t;
          h++;
          *t = c;
          t--;
        }
    }
}

/*
 * ltostr()
 * value - value to be convert to sting
 * ret_str - the converted sting
 * This function will convert the interge to 
 * a string value. The converted value will be stored
 * in ret_str.
 */
static void
ltostr(long value, GR_CHAR **ret_str)
{
        GR_CHAR strval[20];
        int i=1;
        int val;
        int remain;

        remain = value %10;
        val = value/10;

        strval[0] = (GR_CHAR)(remain + '0');
        while(val)
        {
                remain = val %10;
                val = val/10;
                strval[i] = (GR_CHAR)(remain + '0');
                i++;
        }
        strval[i] = 0;
        strreverse(strval);

        if( !(*ret_str = (GR_CHAR*)malloc(i+1)) )
                return;

        strcpy(*ret_str, strval);
}

#else
#define GsDeactivateWin()	print("Nanox-Warning : Only for EmLinux Desktop\n")	
#define GsSendContSig(wp)	print("Nanox-Warning : Only for EmLinux Desktop\n")	
#endif

#endif // NANOGTK

