/*
 * Copyright (c) 1999 Greg Haerr <greg@censoft.com>
 *
 * Nano-X Core Protocol Client Request Handling Routines
 */ 
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <assert.h>
#include <errno.h>
#include <string.h>
#include "serv.h"
#include "nxproto.h"

#define SZREQBUF	2048	/* initial request buffer size*/

#ifndef __ECOS
static REQBUF	reqbuf;		/* request buffer*/
extern int 	nxSocket;
extern char *	nxSharedMem;
#endif

/* Allocate a request buffer of passed size and fill in header fields*/
void *
nxAllocReq(int type, long size, long extra)
{
	nxReq *	req;
	long	aligned_len;
        ACCESS_PER_THREAD_DATA();

	/* variable size requests must be hand-padded*/
	if(extra)
		assert((size & (long)(ALIGNSZ-1)) == 0);

	/* calculate aligned length of request buffer*/
	aligned_len = (size + extra + (long)(ALIGNSZ-1)) & ~(long)(ALIGNSZ-1);

	/* verify we're not greater than max request size*/
	assert(aligned_len <= MAXREQUESTSZ);

	/* flush buffer if required, and allocate larger one if required*/
	if(reqbuf.bufptr + aligned_len >= reqbuf.bufmax)
		nxFlushReq(aligned_len,1);

	/* fill in request header*/
	req = (nxReq *)reqbuf.bufptr;
	req->reqType = (BYTE8)type;
	req->hilength = (BYTE8)((size + extra) >> 16);
	req->length = (UINT16)(size + extra);
	reqbuf.bufptr += aligned_len;

#if 0
{
   struct timeval clienttv;
   struct timezone clienttz;
    
   gettimeofday(&clienttv,&clienttz);
 
   printf("nxAllocReq: [%ld] type=%d size=%d extra=%d aligned_len=%d",
          clienttv.tv_usec, type, size, extra, aligned_len);

   printf("\n");
}
#endif

	return req;
}

static void nxAllocReqbuffer(long newsize)
{
        ACCESS_PER_THREAD_DATA();

	if(newsize < (long)SZREQBUF)
		newsize = SZREQBUF;
	reqbuf.buffer = malloc(newsize);
	if(!reqbuf.buffer) {
		EPRINTF("nxFlushReq: Can't allocate initial request buffer\n");
		exit(1);
	}
	reqbuf.bufptr = reqbuf.buffer;
	reqbuf.bufmax = reqbuf.buffer + newsize;
}

void
nxAssignReqbuffer(char *buffer, long size)
{
        ACCESS_PER_THREAD_DATA();

	if ( reqbuf.buffer != 0 )
		free(reqbuf.buffer);
	reqbuf.buffer = buffer;
	reqbuf.bufptr = reqbuf.buffer;
	reqbuf.bufmax = reqbuf.buffer + size;
}

/* Write a block of data on the socket to the nano-X server */
void
nxWriteSocket(char *buf, int todo)
{
	int written;
        ACCESS_PER_THREAD_DATA();

#if 0
{
   struct timeval clienttv;
   struct timezone clienttz;
   char *s;
   int c, i, j, m;

   s = (char *)buf;
   c = todo;
   m = 16;
   (c <= m) ? (j = c) : (j = m);

   gettimeofday(&clienttv,&clienttz);

   printf("nxWriteSocket: [%ld] fd=%d c=%d ", clienttv.tv_usec, nxSocket, c);

   for (i = 0; i < j; i++)
      printf(" %u", (unsigned char)s[i]);

   printf("\n");
}
#endif

	do {
		written = write(nxSocket, buf, todo);
		if ( written < 0 ) {
			if ( errno == EAGAIN || errno == EINTR )
				continue;
			EPRINTF("nxFlushReq: write failed: %m\n");
			exit(1);
		}
		buf += written;
		todo -= written;
	} while ( todo > 0 );
}

/* Flush request buffer if required, possibly reallocate buffer size*/
void
nxFlushReq(long newsize, int reply_needed)
{
        ACCESS_PER_THREAD_DATA();

	/* handle one-time initialization case*/
	if(reqbuf.buffer == NULL) {
		nxAllocReqbuffer(newsize);
		return;
	}

	/* flush buffer if required*/
	if(reqbuf.bufptr > reqbuf.buffer) {
		char *	buf = reqbuf.buffer;
		int	todo = reqbuf.bufptr - reqbuf.buffer;

#if HAVE_SHAREDMEM_SUPPORT
		if ( nxSharedMem != 0 ) {
			/* There is a shared memory segment used for the
			 * request buffer.  Make up a flush command and
			 * send it over the socket, to tell the server to
			 * process the shared memory segment.
			 * The 'reply_needed' argument should be non-zero
			 * when a confirmation is needed that all commands
			 * are flushed, so new ones can be filled into the
			 * request buffer.  NOTE:  This is *only* needed
			 * when explicitely flushing the request buffer, or
			 * when flushing it to make space for new commands.
			 * DO NOT REQUEST A REPLY when flushing the request
			 * buffer because the last command in the buffer
			 * will send a response:  This response would be
			 * queued up first and had to be drained before the
			 * response to the flush command itsel....
			 * So the GrReadBlock used to read replys to commands
			 * must not specify a nonzero 'reply_needed'.
			 * Not requesting a reply in this case is
			 * safe, since the command executed will wait for
			 * the reply *it* is waiting for, and thus make
			 * sure the request buffer is flushed before
			 * continuing.
			 *
			 * We have to make the protocol request by hand,
			 * as it has to be sent over the socket to wake
			 * up the Nano-X server.
			 */
#ifdef NANOGTK
                        short c;
                        nxShmCmdsFlushReq req;

                        req.reqType = GrNumShmCmdsFlush;
                        req.hilength = 0;
                        req.length = sizeof(req);
                        req.size = todo;
                        req.reply = reply_needed;

//printf("nxFlushReq: (1) todo=%d reply_needed=%d\n", todo, reply_needed);

                        nxWriteSocket((char *)&req,sizeof(req));

                        if (reply_needed)
                        {
                                int e, n, size = 2;
                                char *buf = (char *)(&c);

                                while (1)
                                {
                                        n = 0;
                                        while(n < size) {
                                                e = read(nxSocket,
							((char *) buf + n),
							(size - n));
                                                n += e;
                                        }
                                        if (c==1234)
                                        {
                                                break;
                                        }
                                        else
                                        /* Maybe an event send bye the server */
                                        {
                                                if (c == GrNumGetNextEvent) {
                                                        GR_EVENT event;

	// must not call GrReadBlock() to get event data since it will
	// call nxFlushReq() and hence reqbuf.bufptr is not being reset
	// properly.

        int i = 0;
        char *v, *b;

        b = (char *)&event;
        v = (char *) b;
	n = sizeof(GR_EVENT);

        while(v < ((char *) b + n)) {
                i = read(nxSocket, v, ((char *) b + n - v));
                if ( i <= 0 ) {
                        if ( i == 0 ) {
                                /* We should maybe produce an event here,
                                 * if possible.
                                 */
                                EPRINTF("nxclient: lost connection to Nano-X "
                                        "server\n");
                                exit(1);
                        }
                        if ( errno == EINTR || errno == EAGAIN )
                                continue;

                        EPRINTF("nxclient: bad readblock %d\n", i);
                        return;
                }
                v += i;
        }
                                                        QueueEvent(&event);
                                                }
                                        }
                                }
                        }
#else
			char c;
			nxShmCmdsFlushReq req;

			req.reqType = GrNumShmCmdsFlush;
			req.hilength = 0;
			req.length = sizeof(req);
			req.size = todo;
			req.reply = reply_needed;

			nxWriteSocket((char *)&req,sizeof(req));

			if ( reply_needed )
				while ( read(nxSocket, &c, 1) != 1 )
					;
#endif // NANOGTK
			reqbuf.bufptr = reqbuf.buffer;

			if ( reqbuf.buffer + newsize > reqbuf.bufmax ) {
				/* Shared memory too small, critical */
				EPRINTF("nxFlushReq: shm region too small\n");
				exit(1);
			}
			return;
		}
#endif /* HAVE_SHAREDMEM_SUPPORT*/

		/* Standard Socket transfer */
//printf("nxFlushReq: (2) todo=%d\n", todo);
		nxWriteSocket(buf,todo);
		reqbuf.bufptr = reqbuf.buffer;
	}

	/* allocate larger buffer for current request, if needed*/
	if(reqbuf.bufptr + newsize >= reqbuf.bufmax) {
		reqbuf.buffer = realloc(reqbuf.buffer, newsize);
		if(!reqbuf.buffer) {
		       EPRINTF("nxFlushReq: Can't reallocate request buffer\n");
			exit(1);
		}
		reqbuf.bufptr = reqbuf.buffer;
		reqbuf.bufmax = reqbuf.buffer + newsize;
	}
}

/* calc # bytes required for passed string according to encoding*/
int
nxCalcStringBytes(void *str, int count, int flags)
{
	int	nbytes;

	/* calc byte length of data*/
	if(flags & MWTF_UC16)
		nbytes = count * 2;
	else if(flags & MWTF_UC32)
		nbytes = count * 4;
	else
		nbytes = count;

	return nbytes;
}
