#ifndef srvfunc_gtkC_c
#define srvfunc_gtkC_c

GR_MASK * gr_mask;

#include "cjk.h"

/*
 * draw simple (width = 1) dash line with the segments given
 * only even segments are drawn
 */
static void
draw_dash_line (GR_DRAWABLE *dp, GR_SEGMENT *segment, GR_COUNT n, GR_SIZE line_width)
{
        int i;
        GR_POINT corners[4];
        if (line_width == 1) {
                for (i = 0; i < n; i+=2) {
                        GdLine(dp->psd, dp->x + segment[i].x1, dp->y + segment[i].y1,
                                        dp->x + segment[i].x2, dp->y + segment[i].y2, TRUE);

                }
        }
        else if (line_width > 1) {
                for (i = 0; i < n; i+=2) {
                        calc_thick_line_corners(
                                        (GR_POINT){segment[i].x1, segment[i].y1},
                                        (GR_POINT){segment[i].x2, segment[i].y2},
                                        line_width, corners);
                        draw_thick_line(dp, corners);
                }
        }
}

/*
 * returns an intersection of 2 lines
 *   point pair(1) and point_pair(2)
 *
 */
static void
intersect_line(GR_POINT *line1, GR_POINT *line2, GR_POINT *intersect)
{
        GR_BOOL found_a = GR_FALSE;
        GR_BOOL found_b = GR_FALSE;
        double  a,b;
        double  s1 = 0, s2 = 0;

        if (line1[0].x == line1[1].x) {
                intersect->x = line1[0].x;
                found_a = GR_TRUE;
        }
        else
                s1 = (double)(line1[1].y - line1[0].y) / (double)(line1[1].x - line1[0].x);
        if (line2[0].x == line2[1].x) {
                intersect->x = line2[0].x;
                found_a = GR_TRUE;
        }
        else
                s2 = (double)(line2[1].y - line2[0].y) / (double)(line2[1].x - line2[0].x);
        if (line1[0].y == line1[1].y) {
                intersect->y = line1[0].y;
                found_b = GR_TRUE;
        }
        if (line2[0].y == line2[1].y) {
                intersect->y = line2[0].y;
                found_b = GR_TRUE;
        }
        if (found_a && found_b) {
                return;
        }
        if (found_a && !found_b) {
                a = intersect->x;
                if (intersect->x == line1[0].x)
                        b = s2 * (a - (double)line2[0].x) + (double)line2[0].y;
                else
                        b = s1 * (a - (double)line1[0].x) + (double)line1[0].y;
                intersect->y = b + ((b > 0.0)? 0.5 : -0.5);
                return;
        }
        if (!found_a && found_b) {
                b = intersect->y;
                if (intersect->y == line1[0].y)
                        a = (b - (double)line2[0].y + s2 * (double)line2[0].x) / s2;
                else
                        a = (b - (double)line1[0].y + s1 * (double)line1[0].x) / s1;
                intersect->x = a + ((a > 0.0)? 0.5 : -0.5);
                return;
        }
        a = ((double)line2[0].y - (double)line1[0].y + s1 * (double)line1[0].x - s2 * (double)line2[0].x) / (s1 - s2);
        b = s1 * (a - (double)line1[0].x) + (double)line1[0].y;
        intersect->x = a + ((a > 0.0)? 0.5 : -0.5);
        intersect->y = b + ((b > 0.0)? 0.5 : -0.5);
        return;
}

/*
 * Calculate the offsets in the endpoints of straight line
 * (x1,y1) --> (x2,y2), if it's being displaced by 'parallel_dist'.
 * parallel_dist < 0 ===> leftward (or upward)
 * parallel_dist > 0 ===> rightward (or downward)
 */
static void
line_endpoints_offsets(GR_COORD x1, GR_COORD y1, GR_COORD x2,
                       GR_COORD y2, double length,
                       GR_COORD parallel_dist,
                       GR_COORD *offset_x, GR_COORD *offset_y
                      )
{
        GR_COORD xdelta = x2 - x1;
        GR_COORD ydelta = y2 - y1;
        GR_COORD abs_xdelta = (xdelta > 0)? xdelta : -xdelta;
        GR_COORD abs_ydelta = (ydelta > 0)? ydelta : -ydelta;
        double  roundup = (parallel_dist > 0)? 0.5 : -0.5;
        if (xdelta == 0) {
            *offset_x = parallel_dist;
            *offset_y = 0;
            return;
        }
        if (ydelta == 0) {
            *offset_x = 0;
            *offset_y = parallel_dist;
            return;
        }
        /*
         *   offset_y / parallel_dist = delta_x / length
         *   offset_x / parallel_dist = delta_y / length
         */
        if (((ydelta > 0) && (xdelta > 0))
           || ((ydelta < 0) && (xdelta < 0)) ) {
           *offset_x = ((double)parallel_dist * (double)abs_ydelta / (double)length)  + roundup;
           *offset_y = -(((double)parallel_dist * (double)abs_xdelta / (double)length)  + roundup);
        }
        else {
           *offset_x = ((double)parallel_dist * (double)abs_ydelta / (double)length)  + roundup;
           *offset_y = ((double)parallel_dist * (double)abs_xdelta / (double)length)  + roundup;

        }
}

static int
getVisibleRect(GR_WINDOW *wp, MWRECT **visible, GR_SIZE *numRects)
{
        int count;
        extern MWCLIPREGION *clipregion;

        switch(GdClipArea(wp->psd, wp->x, wp->y, wp->x+wp->width-1, wp->y+wp->height-1))
        {
                case CLIP_VISIBLE:
                        return CLIP_VISIBLE;

                case CLIP_INVISIBLE:
                        return CLIP_INVISIBLE;
        }

        count = clipregion->numRects;
        if(!*visible) *visible =  (MWRECT*)malloc(sizeof(MWRECT)*count);

        if(count)
                memcpy(*visible, clipregion->rects, sizeof(MWRECT)*count);
        *numRects = count;
        return CLIP_PARTIAL;

}

void GrNanoGTKUtilForResize(GR_WINDOW *wp,MWRECT *visibleRect_before, int numRects_before)
{
        MWRECT *visibleRect_after=NULL;
        int numRects_after;
        MWCLIPREGION *diff_region;
        MWRECT *expose_rect;
        MWRECT window_rect={wp->x,wp->y,wp->x+wp->width,wp->y+wp->height};
        int clip;
        GR_WINDOW *sibwp = wp;
        int i;


        if (!numRects_before) return;

        diff_region = GdAllocRegion();

        while (numRects_before-->0)
        {
                GdUnionRectWithRegion(visibleRect_before++, diff_region);
        }
        /* Find out the visible window before resizing */
        GsSetClipWindow(wp, NULL, GR_INCLUDE_INFERIORS);
        clip = getVisibleRect(wp, &visibleRect_after, &numRects_after);
        if (clip ==  CLIP_VISIBLE)
        {
                numRects_after = 1;
                visibleRect_after = &window_rect;
        }
        else
        if (clip == CLIP_INVISIBLE)
                numRects_after = 0;

        while (numRects_after-->0)
        {
                GdSubtractRectFromRegion(visibleRect_after++, diff_region);
        }

        expose_rect = diff_region->rects;
        for (i=0;i<diff_region->numRects; i++)
        {
                //GsWpClearWindow(wp->parent, expose_rect->left - wp->parent->x, expose_rect->top - wp->parent->y, expose_rect->right - expose_rect->left , expose_rect->bottom - expose_rect->top , GR_TRUE);
                GsExposeArea(wp->parent, expose_rect->left , expose_rect->top , expose_rect->right - expose_rect->left , expose_rect->bottom - expose_rect->top, NULL);
                expose_rect++;
        }

#if 0
        while (sibwp->siblings)
        {
                expose_rect = diff_region->rects;
                sibwp = sibwp->siblings;
                if (sibwp->output==GR_TRUE)
                {
                        GsPrepareWindow(sibwp->id);
                        for (i=0;i<diff_region->numRects; i++)
                        {
                                GsExposeArea(sibwp, expose_rect->left - wp->parent->x, expose_rect->top - wp->parent->y, expose_rect->right - expose_rect->left , expose_rect->bottom - expose_rect->top );
                                expose_rect++;
                        }
                }
        }
#endif
        GdDestroyRegion(diff_region);

}

void GrNanoGTKResizeWindow(GR_WINDOW *wp, GR_COORD width, GR_COORD height)
{
        int numRects_before;
        MWRECT *visibleRect_before=NULL;
        MWRECT window_rect={wp->x,wp->y,wp->x+wp->width,wp->y+wp->height};
        int clip;


        /* Find out the visible window before resizing */
        GsSetClipWindow(wp, NULL, GR_INCLUDE_INFERIORS);
        clip = getVisibleRect(wp, &visibleRect_before, &numRects_before);
        if (clip ==  CLIP_VISIBLE)
        {
                numRects_before = 1;
                visibleRect_before = &window_rect;
        }
        else
        if (clip == CLIP_INVISIBLE)
                numRects_before = 0;

        if (width < wp->width)
        {
                if (height <= wp->height)
                {
                        wp->width = width;
                        wp->height = height;
                        GrNanoGTKUtilForResize(wp,visibleRect_before,numRects_before);
                }
                else if  (height > wp->height)
                {
                        int old_height = wp->height;
                        /* Expose our parent on the right and ourself on the bottom */
                        wp->width = width;
                        GrNanoGTKUtilForResize(wp,visibleRect_before,numRects_before);
                        wp->height = height;
                        //GsWpClearWindow(wp, 0, old_height, wp->width , height - old_height , GR_TRUE);
                        GsExposeArea(wp, wp->x + 0, wp->y + old_height, wp->width , height - old_height, NULL);
                }
                else
                {
                        /* Else, only for width */
                        return;
                }
        }
        else if (width>=wp->width)
        {
                if (height<wp->height)
                {
                        int old_width = wp->width;
                        /* Expose our parent on the bottom and ourself on the right */
                        wp->height = height;
                        GrNanoGTKUtilForResize(wp,visibleRect_before,numRects_before);
                        if (width!=wp->width)
                        {
                                wp->width = width;
                                //GsWpClearWindow(wp, old_width, 0, width -old_width , height , GR_TRUE);
                                GsExposeArea(wp, wp->x + old_width, wp->y + 0, width -old_width , height, NULL);
                        }
                }
                else
                if (height >= wp->height)
                {
                        int old_width = wp->width;
                        int old_height = wp->height;
                        wp->width = width;
                        wp->height = height;
                        if (old_width!=wp->width)
                        {
                                //GsWpClearWindow(wp, old_width, 0, width-old_width , height, GR_TRUE);
                                GsExposeArea(wp, wp->x + old_width, wp->y + 0, width-old_width , height, NULL);
                        }
                        if (old_height!=wp->height)
                        {
                                //GsWpClearWindow(wp, 0, old_height, old_width , height - old_height, GR_TRUE);
                                GsExposeArea(wp, wp->x, wp->y + old_height, old_width , height - old_height, NULL);
                        }
                }
                else
                        return;
        }
}


/*
 * return 4 corners of a thick line from p1 -> p2.
 * the 4 corners returned are  suitable to just pass into GdFillPoly
 * as is.
 */
static void
calc_thick_line_corners (GR_POINT p1, GR_POINT p2,
                                                 int line_width, GR_POINT *corners)
{
        double  line_len;
        int             start_offset, end_offset;
        int             x_offset, y_offset;

//        assert (line_width > 1);
        /*
         * get the 4 points of the rectangle,
         * and draw it as a filled poly
         */
        start_offset = line_width / 2;
        end_offset = start_offset - line_width + 1;
        line_len = sqrt ((p2.x - p1.x)*(p2.x - p1.x) + (p2.y - p1.y)*(p2.y - p1.y));
        line_endpoints_offsets(p1.x, p1.y, p2.x, p2.y, line_len,
                                                        start_offset, &x_offset, &y_offset);
        corners[0].x = p1.x + x_offset;
        corners[0].y = p1.y + y_offset;
        corners[1].x = p2.x + x_offset;
        corners[1].y = p2.y + y_offset;
        line_endpoints_offsets(p1.x, p1.y, p2.x, p2.y, line_len,
                                                        end_offset, &x_offset, &y_offset);
        corners[2].x = p2.x + x_offset;
        corners[2].y = p2.y + y_offset;
        corners[3].x = p1.x + x_offset;
        corners[3].y = p1.y + y_offset;
}

/*
 * draw a thick line based on the 4 corners, assuming they are
 * in the order suitable for GdFillPoly
 */
static void
draw_thick_line (GR_DRAWABLE *dp, GR_POINT *corners)
{
        int i;

        for (i = 0; i < 4; i++) {
                corners[i].x += dp->x;
                corners[i].y += dp->y;
    }
        GdFillPoly(dp->psd, 4, corners);
}

/*
 * Return segments of a dashed line based on starting and
 * ending co-ordinates, together dash-related info.
 * Assume x1,x2,y1,y2 are all >= 0
 */
static void
calc_dash_line_coords (GR_COORD x1, GR_COORD y1,
                                        GR_COORD x2, GR_COORD y2,
                                        GR_DASH_SIZE *dash_lens, GR_COUNT n_dash_lens,
                                        GR_SEGMENT **psegments, int *n_segments){
        double line_len;
        double rem_line_len;
        int      cnt, d_idx;
        GR_SEGMENT *seg, *prev_seg;
        GR_BOOL incr_x = x2 > x1;
        GR_BOOL incr_y = y2 > y1;
        GR_BOOL done = FALSE;

        /* allocate segments */
        line_len = sqrt((y2 - y1)*(y2 - y1) + (x2 - x1)*(x2 - x1));
        for (cnt = 0, d_idx = 0, rem_line_len = line_len;
                        rem_line_len > 0.0;
                        rem_line_len -= dash_lens[d_idx%n_dash_lens], cnt++, d_idx++) {
                /* just do it inside the loop */
        }
        *n_segments = cnt;
        *psegments = (GR_SEGMENT*)malloc(cnt * sizeof(GR_SEGMENT));
        prev_seg = NULL;
        seg = *psegments;
        seg->x1 = x1;
        seg->y1 = y1;
        for (d_idx= 0; d_idx < cnt && !done;
                d_idx++, prev_seg = seg, seg++) {
                if (prev_seg != NULL) { seg->x1 = prev_seg->x2; }
                if (prev_seg != NULL) { seg->y1 = prev_seg->y2; }
                /* calculate offset */
                seg->x2 = seg->x1 + (double)((x2 - x1) * (double)dash_lens[d_idx%n_dash_lens] / line_len + 0.5);
                seg->y2 = seg->y1 + (double)((y2 - y1) * (double)dash_lens[d_idx%n_dash_lens] / line_len + 0.5);
                if (incr_x) {
            if (seg->x2 >= x2) {
                                seg->x2 = x2;
                                done = GR_TRUE;;
                        }
                }
                else if (x1 != x2) {
                        if (seg->x2 <= x2) {
                                seg->x2 = x2;
                                done = GR_TRUE;;
                        }
                }
                if (incr_y) {
                        if (seg->y2 >= y2) {
                                seg->y2 = y2;
                                done = GR_TRUE;;
                        }
                }
                else if (y1 != y2) {
                        if      (seg->y2 <= y2) {
                                seg->y2 = y2;
                                done = GR_TRUE;;
                        }
                }
        }
        *n_segments = d_idx;
}

/*
 * Return 4 (one interpolated)joining points based on the 2 lines with
 * 4 points each (1 for each corner because of line width > 1)
 */
static void
find_join_points(GR_POINT *left_p1_line1, GR_POINT *right_p1_line1,
                                GR_POINT *left_p2_line1, GR_POINT *right_p2_line1,
                                GR_POINT *left_p1_line2, GR_POINT *right_p1_line2,
                                GR_POINT *left_p2_line2, GR_POINT *right_p2_line2,
                                GR_POINT *join_pts)
{
        int i;
        GR_POINT line1[2], line2[2];
        GR_POINT intersect[4][4] = {
                        {*left_p1_line1, *left_p2_line1, *left_p1_line2, *left_p2_line2 },
                        {*left_p1_line1, *left_p2_line1, *right_p1_line2, *right_p2_line2 },
                        {*right_p1_line1, *right_p2_line1, *right_p1_line2, *right_p2_line2 },
                        {*right_p1_line1, *right_p2_line1, *left_p1_line2, *left_p2_line2 },
                        };
        /* calculate the join points one-by-one */
        for (i = 0; i < 4; i++) {
                line1[0] = intersect[i][0];
                line1[1] = intersect[i][1];
                line2[0] = intersect[i][2];
                line2[1] = intersect[i][3];
                intersect_line (line1, line2, &(join_pts[i]));
        }
}

/*
 * add dp->x,y  to points
 */
static void
add_dp_offset (GR_DRAWABLE *dp, GR_POINT *points, GR_COUNT n)
{
        int i;
        for (i = 0; i < n; i++) {
                        points[i].x += dp->x;
                        points[i].y += dp->y;
        }
}

static void
updateChildCursor(GR_WINDOW *wp)
{
        GR_WINDOW *childwp;

        for(childwp = wp->children; childwp; childwp = childwp->siblings)
        {
                if(childwp->useparent && childwp->cursor != wp->cursor)
                {
                        childwp->cursor->usecount --;
                        if (childwp->cursor->usecount == 0)
                                free(childwp->cursor);                           
                        childwp->cursor = wp->cursor;
                        wp->cursor->usecount++;
                        updateChildCursor(childwp);
                }
        }
}

/*
 * Specify a cursor for a window.
 * This cursor will only be used within that window, and by default
 * for its new children.  If the cursor is currently within this
 * window, it will be changed to the new one immediately.
 */
void
GrSetCursor_aux(GR_WINDOW_ID wid, GR_SIZE width, GR_SIZE height, GR_COORD hotx,
        GR_COORD hoty, GR_COLOR foreground, GR_COLOR background, int useparent,
        GR_BITMAP *fgbitmap, GR_BITMAP *bgbitmap)
{
        GR_WINDOW       *wp;            /* window structure */
        GR_CURSOR       *cp;            /* cursor structure */
        int             bytes;          /* number of bytes of bitmap */
        int             new = GR_FALSE;

        wp = GsFindWindow(wid);
        if (wp == NULL)
                return;

        if(useparent)
        {
                wp->useparent = 1;
                return;
        }

        /*
         * Make sure the size of the bitmap is reasonable.
         */
        if ((width <= 0) || (width > MWMAX_CURSOR_SIZE) ||
                (height <= 0) || (height > MWMAX_CURSOR_SIZE))
        {
                GsError(GR_ERROR_BAD_CURSOR_SIZE, 0);
                return;
        }
        bytes = GR_BITMAP_SIZE(width, height) * sizeof(GR_BITMAP);

        /*
         * See if the window is using a shared cursor definition.
         * If so, then allocate a new private one, otherwise reuse it.
         */
        cp = wp->cursor;
        if (cp == NULL || cp->usecount-- > 1) {
                cp = (GR_CURSOR *) malloc(sizeof(GR_CURSOR));
                if (cp == NULL) {
                        GsError(GR_ERROR_MALLOC_FAILED, 0);
                        return;
                }
                new = GR_TRUE;
        }

        cp->usecount = 1;
        cp->cursor.width = width;
        cp->cursor.height = height;
        cp->cursor.hotx = hotx;
        cp->cursor.hoty = hoty;
        cp->cursor.fgcolor = foreground;
        cp->cursor.bgcolor = background;
        memcpy(cp->cursor.image, fgbitmap, bytes);
        memcpy(cp->cursor.mask, bgbitmap, bytes);
        wp->cursor = cp;
        wp->useparent = 0;
        /*
         * Now, update all children cursor as well, however, only update
         * those children with useparent flag is set.
         */
        if(new) updateChildCursor(wp);

        /*
         * If this was the current cursor, then draw the new one.
         */
        if (cp == curcursor || curcursor == NULL) {
                GdMoveCursor(cursorx - cp->cursor.hotx,
                        cursory - cp->cursor.hoty);
                GdSetCursor(&cp->cursor);
        }

        GsCheckCursor();
}

void
GrAddClipRectangles(GR_GC_ID gcid, GR_COORD origin_x, GR_COORD origin_y, GR_SIZE n, GR_RECT *cliprect)
{
        GR_GC *gcp;
        MWRECT rc;

        gcp = GsFindGC(gcid);
        if(!gcp)
                return;
        if(n > 0)
        {
                GR_REGION *regionp= NULL;
                int i;
                if(gcp->regionid)
                {
                        regionp = GsFindRegion(gcp->regionid);
                        if(!regionp)
                                goto createregion;
                }
                else
                {
createregion:           /* Create a new Region */
                        regionp = (GR_REGION *) malloc(sizeof(GR_REGION));
                        if (regionp == NULL)
                        {
                                GsError(GR_ERROR_MALLOC_FAILED, 0);
                                return;
                        }
                        regionp->rgn = GdAllocRegion();
                        regionp->id = nextregionid++;
                        regionp->next = listregionp;
                        regionp->owner = curclient;
                        listregionp = regionp;
                }
                for (i=0;i<n;i++)
                {
                        rc.left = cliprect[i].x;
                        rc.top = cliprect[i].y;
                        rc.right = cliprect[i].x + cliprect[i].width;
                        rc.bottom = cliprect[i].y + cliprect[i].height;
                        GdUnionRectWithRegion(&rc,regionp->rgn);
                }
                gcp->regionid = regionp->id;
                gcp->changed = GR_TRUE;
                return;
        }
}

/*
 * Set the subwindow mode which determines how drawing onto a window
 * will affect child windows of that window.
 */
void
GrSetGCSubWindowMode(GR_GC_ID gc, int subwindowmode)
{
        GR_GC           *gcp;           /* graphics context */

        gcp = GsFindGC(gc);
        if (!gcp || gcp->subwindowmode == subwindowmode)
                return;

        gcp->subwindowmode = subwindowmode;
        gcp->changed = GR_TRUE;
}

#ifdef LINE_ATTRIB_SUPPORT
/*
 * Set the line-attrib for drawing under this graphics context
 */
void
GrSetGCLineAttrib(GR_GC_ID gc, GR_LINE_ATTRIB_MASK value_mask,
                  GR_SIZE line_width, GR_LINE_STYLE line_style)
{
        GR_GC           *gcp;           /* graphics context */

        gcp = GsFindGC(gc);
        if (!gcp)
                return;
        if (value_mask & GR_LINE_ATTRIB_WIDTH_SET)
                gcp->line_width = line_width;
        if (value_mask & GR_LINE_ATTRIB_STYLE_SET)
                gcp->line_style = line_style;
        if (value_mask & (GR_LINE_ATTRIB_WIDTH_SET | GR_LINE_ATTRIB_STYLE_SET))
                gcp->changed = GR_TRUE;
}

/*
 * Set the line dash sizes for drawing under this graphics context
 */
void
GrSetGCLineDashSizes(GR_GC_ID gc, GR_COUNT n, GR_DASH_SIZE *dashes)
{
        GR_GC           *gcp;           /* graphics context */

        gcp = GsFindGC(gc);
        if (!gcp)
                return;
        gcp->n_dashes = n;
        gcp->dash_list = (GR_DASH_SIZE*)malloc(n * sizeof(GR_DASH_SIZE));
        memcpy (gcp->dash_list, dashes, n * sizeof(GR_DASH_SIZE));
        gcp->changed = GR_TRUE;
}
#endif /* LINE_ATTRIB_SUPPORT */

void
GrExitDrawingMode()
{
#ifdef HAVE_HWR_SUPPORT
        hwr_leave_draw_mode();
#endif
}

void
GrSetGCRegionX(GR_GC_ID gcid, int size, int numRects, int sub_total, MWRECT *extents, MWRECT * rects)
{
        GR_GC *gcp;

        gcp = GsFindGC(gcid);
        if(!gcp)
                return;

        if(numRects > 0)
        {
                GR_REGION *regionp= NULL;
                MWCLIPREGION * region;

                if(gcp->regionid)
                {
                        regionp = GsFindRegion(gcp->regionid);
                        if(!regionp)
                                goto createregion;
                        GdEmptyRegion(regionp->rgn);
                }
                else
                {
createregion:           /* Create a new Region */
                        regionp = (GR_REGION *) malloc(sizeof(GR_REGION));
                        if (regionp == NULL)
                        {
                                GsError(GR_ERROR_MALLOC_FAILED, 0);
                                return;
                        }
                        regionp->rgn = GdAllocRegion();
                        regionp->id = nextregionid++;
                        regionp->next = listregionp;
                        regionp->owner = curclient;
                        listregionp = regionp;
                }
                region = regionp->rgn;
                region->size = size;
                region->numRects = sub_total;
                region->type = MWREGION_COMPLEX; //?
                memcpy(&region->extents,extents,sizeof(MWRECT));
                region->rects = (MWRECT *)realloc(region->rects,numRects*sizeof(MWRECT));
                memcpy(region->rects,rects,sub_total*sizeof(MWRECT));

                gcp->regionid = regionp->id;
                gcp->changed = GR_TRUE;
                return;
        }
}

void
GrAddGCRegionX(GR_GC_ID gcid, int size, int numRects, int sub_total, MWRECT * rects)
{
        GR_GC *gcp;

        gcp = GsFindGC(gcid);
        if(!gcp)
                return;

        if(numRects > 0)
        {
                GR_REGION *regionp= NULL;
                MWCLIPREGION * region;

//                assert(gcp->regionid);
                regionp = GsFindRegion(gcp->regionid);
//                assert(regionp);
                region = regionp->rgn;
//                assert(region->size==size);
                memcpy(region->rects+region->numRects,rects,sub_total*sizeof(MWRECT));
                region->numRects+=sub_total;

                gcp->regionid = regionp->id;
                gcp->changed = GR_TRUE;
                return;
        }
}

void
GrSetBackgroundColor(GR_WINDOW_ID wid, GR_COLOR color)
{
        GR_WINDOW       *wp;

        wp = GsFindWindow(wid);

        if (!wp || wp->background == color )
                        return;

       wp->background = color;
}

void
GrGrabPointer(GR_WINDOW_ID wid, GR_EVENT_MASK mask) {
#ifdef DEBUG
        printf("GrGrabPointer %d %d\n", wid, mask);
#endif

        if (wid==0) {  // wid should not be 0
                win_grabbed_ptr_mask=0;
        win_grabbed_ptr=NULL;
            return;
        }

    if ((win_grabbed_ptr = GsFindWindow(wid))==NULL) {
                printf("GrGrabPointer cannot find window with id:%d\n", wid);
        }

        /* FIXME: should work for all kind of events, but now only
         *        GR_EVENT_MASK_BUTTON_DOWN and GR_EVENT_MASK_BUTTON_UP */
        if (mask & (~(GR_EVENT_MASK_BUTTON_DOWN|GR_EVENT_MASK_BUTTON_UP))) {
                printf("GrGrabPointer: Warning: only button up and down supported\n");
        }
        win_grabbed_ptr_mask=mask;
}

void
GrScrollWindow(GR_DRAW_ID wid, GR_GC_ID gc, GR_COORD offx, GR_COORD offy)
{
        GR_DRAWABLE     *dp;
        GR_WINDOW       *wp;
        GR_WINDOW       *childwp;
        GR_DRAW_TYPE    type;
        MWRECT          *exp_rect;
        extern MWCLIPREGION *clipregion;
        int             nb_rect;
        GR_GC           *gcp;
        int             old_gc_mode;

        wp = GsFindWindow(wid);

        gcp = GsFindGC(gc);
        if (!gcp)
                return;

        /* The clip region should include the children */
        old_gc_mode = gcp->subwindowmode;
        if (gcp->subwindowmode != GR_INCLUDE_INFERIORS)
        {
                gcp->subwindowmode = GR_INCLUDE_INFERIORS;
                gcp->changed = GR_TRUE;
        }


        type = GsPrepareDrawing(wid, gc, &dp);

        if ( (!wp ) || type == GR_DRAW_TYPE_NONE)
                return;

#ifndef OLD
        if (offy<0)
        {
                if (offx<0)
                        GdBlit(dp->psd, dp->x - offx, dp->y - offy, dp->width + offx , dp->height + offy,
                                dp->psd, dp->x, dp->y, MWROP_SRCCOPY);
                else
                        GdBlit(dp->psd, dp->x , dp->y - offy, dp->width - offx, dp->height + offy,
                                dp->psd, dp->x + offx, dp->y, MWROP_SRCCOPY);
        }
        else
        {
                if (offx<0)
                        GdBlit(dp->psd, dp->x - offx, dp->y , dp->width + offx , dp->height - offy,
                                dp->psd, dp->x, dp->y + offy, MWROP_SRCCOPY);
                else
                        GdBlit(dp->psd, dp->x , dp->y , dp->width - offx, dp->height - offy,
                                dp->psd, dp->x + offx, dp->y + offy, MWROP_SRCCOPY);
        }
#else
        if (offy<0)
                GdBlit(dp->psd, dp->x , dp->y - offy, dp->width, dp->height + offy,
                        dp->psd, dp->x, dp->y, MWROP_SRCCOPY);
        else
        if (offy>0)
                GdBlit(dp->psd, dp->x , dp->y , dp->width, dp->height - offy,
                        dp->psd, dp->x, dp->y + offy, MWROP_SRCCOPY);
        else
        if (offx<0)
                GdBlit(dp->psd, dp->x - offx, dp->y , dp->width + offx , dp->height,
                        dp->psd, dp->x, dp->y, MWROP_SRCCOPY);
        else
        if (offx>0)
                GdBlit(dp->psd, dp->x , dp->y , dp->width - offx, dp->height,
                        dp->psd, dp->x + offx, dp->y, MWROP_SRCCOPY);
#endif

        /* Move the children */
        for (childwp=wp->children;childwp;childwp=childwp->siblings)
                OffsetWindow(childwp,  - offx,  - offy);


        if ((nb_rect = clipregion->numRects)>=1)
        {
#ifdef OLD
                GR_COORD l_rootx,l_rooty,l_width,l_height;
#endif
                MWCLIPREGION    *whole_region;
                MWCLIPREGION    *clip_region_offset;
                MWCLIPREGION    *whole_region_offset;

                whole_region = GdAllocRectRegion( wp->x, wp->y, wp->x + wp->width,wp->y + wp->height);
                GdSubtractRegion(whole_region, whole_region, clipregion);

#ifndef OLD
                clip_region_offset = GdAllocRegion();
                GdCopyRegion(clip_region_offset, clipregion);
                GdOffsetRegion(clip_region_offset, -offx, -offy);

                whole_region_offset = GdAllocRectRegion( wp->x, wp->y, wp->x + wp->width,wp->y + wp->height);
                GdSubtractRegion(whole_region_offset, whole_region_offset, clip_region_offset);
                GdSubtractRegion(whole_region_offset, whole_region_offset, whole_region);

                exp_rect = whole_region_offset->rects;
                while (whole_region_offset->numRects--)
                {
                        GsExposeAreaOnlyChild(wp, exp_rect->left,exp_rect->top,exp_rect->right - exp_rect->left,exp_rect->bottom - exp_rect->top);
                        exp_rect++;
                }
                GdDestroyRegion(clip_region_offset);
                GdDestroyRegion(whole_region_offset);

#else
#if 0
                /* Up */
                if (offy > 0)
                {
                        if (offx < 0)
                        {
                                /* Up right */
                                GsExposeAreaOnlyChild(wp, wp->x,wp->y+wp->height - offy ,wp->width,offy);
                                GsExposeAreaOnlyChild(wp, wp->x,wp->y ,-offx, wp->height - offy);

                                exp_rect = whole_region->rects;
                                while (whole_region->numRects--)
                                {
                                        l_rootx = exp_rect->left - offx;
                                        l_rooty = exp_rect->top - offy;
                                        l_width = exp_rect->right - exp_rect->left;
                                        l_height = MWMIN(offy,exp_rect->bottom - exp_rect->top);
                                        GsExposeArea(wp, l_rootx,l_rooty,l_width,l_height, NULL);

                                        l_rootx = exp_rect->right;
                                        l_rooty = exp_rect->top;
                                        l_width = MWMIN(-offx,exp_rect->right - exp_rect->left);
                                        l_height = exp_rect->bottom - exp_rect->top - offy;
                                        GsExposeArea(wp, l_rootx,l_rooty,l_width,l_height, NULL);

                                        exp_rect++;
                                }
                        }
                        else
                        {
                                /* Up left */
                                GsExposeAreaOnlyChild(wp, wp->x,wp->y+wp->height - offy ,wp->width,offy);
                                GsExposeAreaOnlyChild(wp, wp->x + wp->width - offx ,wp->y ,offx,wp->height - offy);

                                exp_rect = whole_region->rects;
                                while (whole_region->numRects--)
                                {
                                        l_rootx = exp_rect->left - offx;
                                        l_rooty = exp_rect->top - offy;
                                        l_width = exp_rect->right - exp_rect->left;
                                        l_height = MWMIN(offy,exp_rect->bottom - exp_rect->top);
                                        GsExposeArea(wp, l_rootx,l_rooty,l_width,l_height, NULL);

                                        l_rootx = exp_rect->left - offx;
                                        l_rooty = exp_rect->top;
                                        l_width = MWMIN(offx,exp_rect->right - exp_rect->left);
                                        l_height = exp_rect->bottom - exp_rect->top - offy;
                                        GsExposeArea(wp, l_rootx,l_rooty,l_width,l_height, NULL);

                                        exp_rect++;
                                }
                        }
                }
                /* Down */
                else if (offy < 0)
                {
                        if (offx < 0)
                        {
                                /* Down right */
                                GsExposeAreaOnlyChild(wp, wp->x,wp->y ,wp->width, -offy);
                                GsExposeAreaOnlyChild(wp, wp->x,wp->y-offy ,-offx, wp->height + offy);

                                exp_rect = whole_region->rects;
                                while (whole_region->numRects--)
                                {
                                        l_rootx = exp_rect->left - offx;
                                        l_rooty = exp_rect->bottom;
                                        l_width = exp_rect->right - exp_rect->left;
                                        l_height = MWMIN(-offy,exp_rect->bottom - exp_rect->top);
                                        GsExposeArea(wp, l_rootx,l_rooty,l_width,l_height, NULL);

                                        l_rootx = exp_rect->right;
                                        l_rooty = exp_rect->top - offy;
                                        l_width = MWMIN(-offx,exp_rect->right - exp_rect->left);
                                        l_height = exp_rect->bottom - exp_rect->top;
                                        GsExposeArea(wp, l_rootx,l_rooty,l_width,l_height, NULL);

                                        exp_rect++;
                                }
                        }
                        else
                        {
                                /* Down left */
                                GsExposeAreaOnlyChild(wp, wp->x,wp->y ,wp->width, -offy);
                                GsExposeAreaOnlyChild(wp, wp->x + wp->width - offx ,wp->y-offy ,offx,wp->height + offy);

                                exp_rect = whole_region->rects;
                                while (whole_region->numRects--)
                                {
                                        l_rootx = exp_rect->left;
                                        l_rooty = exp_rect->bottom;
                                        l_width = exp_rect->right - exp_rect->left - offx;
                                        l_height = MWMIN(-offy,exp_rect->bottom - exp_rect->top);
                                        GsExposeArea(wp, l_rootx,l_rooty,l_width,l_height, NULL);

                                        l_rootx = exp_rect->left - offx;
                                        l_rooty = exp_rect->top - offy;
                                        l_width = MWMIN(offx,exp_rect->right - exp_rect->left);
                                        l_height = exp_rect->bottom - exp_rect->top;
                                        GsExposeArea(wp, l_rootx,l_rooty,l_width,l_height, NULL);

                                        exp_rect++;
                                }
                        }
                }
                /* Left */
                else if (offx > 0)
                {
                        GsExposeAreaOnlyChild(wp, wp->x + wp->width - offx ,wp->y ,offx,wp->height);
                        exp_rect = whole_region->rects;
                        while (whole_region->numRects--)
                        {
                                l_rootx = exp_rect->left - offx;
                                l_rooty = exp_rect->top;
                                l_width = offx;
                                l_height = exp_rect->bottom - exp_rect->top;
                                GsExposeArea(wp, l_rootx,l_rooty,l_width,l_height, NULL);
                                exp_rect++;
                        }
                }
                /* Right */
                else if (offx < 0)
                {
                        GsExposeAreaOnlyChild(wp, wp->x,wp->y ,-offx, wp->height);
                        exp_rect = whole_region->rects;
                        while (whole_region->numRects--)
                        {
                                l_rootx = exp_rect->right;
                                l_rooty = exp_rect->top;
                                l_width = -offx ;
                                l_height = exp_rect->bottom - exp_rect->top;
                                GsExposeArea(wp, l_rootx,l_rooty,l_width,l_height, NULL);
                                exp_rect++;
                        }
                }
#else
                /* Up */
                if (offy > 0)
                {
                        GsExposeAreaOnly(wp, wp->x,wp->y+wp->height - offy ,wp->width,offy);
                        exp_rect = whole_region->rects;
                        while (whole_region->numRects--)
                        {
                                l_rootx = exp_rect->left;
                                l_rooty = exp_rect->top - offy;
                                l_width = exp_rect->right - exp_rect->left;
                                l_height = offy;
                                GsExposeArea(wp, l_rootx,l_rooty,l_width,l_height, NULL);
                                exp_rect++;
                        }
                }
                /* Down */
                else if (offy < 0)
                {
                        GsExposeAreaOnly(wp, wp->x,wp->y ,wp->width, -offy);
                        exp_rect = whole_region->rects;
                        while (whole_region->numRects--)
                        {
                                l_rootx = exp_rect->left;
                                l_rooty = exp_rect->bottom;
                                l_width = exp_rect->right - exp_rect->left;
                                l_height = -offy;
                                GsExposeArea(wp, l_rootx,l_rooty,l_width,l_height, NULL);
                                exp_rect++;
                        }
                }
                /* Left */
                else if (offx > 0)
                {
                        GsExposeAreaOnly(wp, wp->x + wp->width - offx ,wp->y ,offx,wp->height);
                        exp_rect = whole_region->rects;
                        while (whole_region->numRects--)
                        {
                                l_rootx = exp_rect->left - offx;
                                l_rooty = exp_rect->top;
                                l_width = offx;
                                l_height = exp_rect->bottom - exp_rect->top;
                                GsExposeArea(wp, l_rootx,l_rooty,l_width,l_height, NULL);
                                exp_rect++;
                        }
                }
                /* Right */
                else if (offx < 0)
                {
                        GsExposeAreaOnly(wp, wp->x,wp->y ,-offx, wp->height);
                        exp_rect = whole_region->rects;
                        while (whole_region->numRects--)
                        {
                                l_rootx = exp_rect->right;
                                l_rooty = exp_rect->top;
                                l_width = -offx ;
                                l_height = exp_rect->bottom - exp_rect->top;
                                GsExposeArea(wp, l_rootx,l_rooty,l_width,l_height, NULL);
                                exp_rect++;
                        }
                }
#endif
#endif
                GdDestroyRegion(whole_region);
        }
        if (gcp->subwindowmode != old_gc_mode)
        {
                gcp->subwindowmode = old_gc_mode;
                gcp->changed = GR_TRUE;
        }
}

int
GrGetNextExposeWin(GR_EVENT *ep, GR_WINDOW_ID wid)
{
        GR_EVENT_LIST * elp;
        GR_EVENT_LIST * prev=NULL;

        elp = curclient->eventhead;

        /* If wid!=-1, return the next event whose type and wid matches */
        while (elp!=NULL)
        {
                if ((elp->event.type == GR_EVENT_TYPE_EXPOSURE) && ( (wid == -1) || (elp->event.exposure.wid == wid) ))
                {
                        *ep = elp->event;
                        if (!prev)
                                curclient->eventhead = elp->next;
                        else
                                prev->next = elp->next;

                        if (curclient->eventtail == elp)
                                curclient->eventtail = prev;


                        /* On the free queue */
                        elp->next = eventfree;
                        eventfree = elp;
                        return 1;
                }
                prev = elp;
                elp=elp->next;
        }
        ep->type = GR_EVENT_TYPE_NONE;
        return 0;
}

void
GrSetClipRectangles(GR_GC_ID gcid, GR_COORD origin_x, GR_COORD origin_y, GR_SIZE n, GR_RECT *cliprect)
{
        GR_GC *gcp;
        MWRECT rc;

        gcp = GsFindGC(gcid);
        if(!gcp)
                return;

        if(n > 0)
        {
                GR_REGION *regionp= NULL;
#if 1
                int i;
                if(gcp->regionid)
                {
                        regionp = GsFindRegion(gcp->regionid);
                        if(!regionp)
                                goto createregion;
                        //GdDestroyRegion(regionp->rgn);
                        GdEmptyRegion(regionp->rgn);
                }
                else
                {
createregion:           /* Create a new Region */
                        regionp = (GR_REGION *) malloc(sizeof(GR_REGION));
                        if (regionp == NULL)
                        {
                                GsError(GR_ERROR_MALLOC_FAILED, 0);
                                return;
                        }
                        regionp->rgn = GdAllocRegion();
                        regionp->id = nextregionid++;
                        regionp->next = listregionp;
                        regionp->owner = curclient;
                        listregionp = regionp;
                }
                for (i=0;i<n;i++)
                {
                        rc.left = cliprect[i].x;
                        rc.top = cliprect[i].y;
                        rc.right = cliprect[i].x + cliprect[i].width;
                        rc.bottom = cliprect[i].y + cliprect[i].height;
                        GdUnionRectWithRegion(&rc,regionp->rgn);
                }
                gcp->regionid = regionp->id;
                gcp->changed = GR_TRUE;
                return;
#else
                rc.left = cliprect->x;
                rc.top = cliprect->y;
                rc.right = cliprect->x + cliprect->width;
                rc.bottom = cliprect->y + cliprect->height;
                if(gcp->regionid)
                {
                        regionp = GsFindRegion(gcp->regionid);
                        if(!regionp)
                                goto createregion;
                }
                else
                {
                createregion:
                        /* Create a new Region */
                        regionp = (GR_REGION *) malloc(sizeof(GR_REGION));
                        if (regionp == NULL)
                        {
                                GsError(GR_ERROR_MALLOC_FAILED, 0);
                                return;
                        }
                        regionp->rgn = GdAllocRegion();
                        regionp->id = nextregionid++;
                        regionp->next = listregionp;
                        regionp->owner = curclient;
                        listregionp = regionp;
                }
                GdSetRectRegion(regionp->rgn, rc.left, rc.top, rc.right, rc.bottom);
                gcp->regionid = regionp->id;
                gcp->changed = GR_TRUE;
                return;
#endif
        }
}

/*
 * If the given wid is not 0, then find the pointer x, y coordinates relative
 * to the given window origin.
 * If the given wid is 0, then find the window that the pointer located in
 * & return the window id & the pointer relative coordinates.
 */
void
GrGetPointer(GR_WINDOW_ID wid, GR_WINDOW_ID *return_wid, GR_COORD *coordx, GR_COORD *coordy)
{
        GR_WINDOW       *wp;

        if(wid)
        {
                wp = GsFindWindow(wid);
                if ((wp == NULL) || (wp->unmapcount ))
                        return ;
        }
        else
        {
                /*      FIXME : r we assume the cursor will always in
                        mousewp at this point.
                        wp = GsFindVisibleWindow(cursorx, cursory);
                 */
                wp = mousewp;
                *return_wid = wp->id;
        }
        *coordx = cursorx - wp->x;
        *coordy = cursory - wp->y;
}

void
GrMoveResizeWindow(GR_WINDOW_ID wid, GR_COORD x, GR_COORD y, GR_SIZE width, GR_SIZE height)
{
        GR_WINDOW       *wp;            /* window structure */
        GR_COORD        offx, offy;

        wp = GsFindWindow(wid);
        if (!wp) return;
        if (wp == rootwp) {
                GsError(GR_ERROR_ILLEGAL_ON_ROOT_WINDOW, wid);
                return;
        }

        x += wp->parent->x;
        y += wp->parent->y;

        if (wp->x == x && wp->y == y && wp->width == width && wp->height == height)
                return;

        if ((width <= 0) || (height <= 0)) {
                GsError(GR_ERROR_BAD_WINDOW_SIZE, wid);
                return;
        }

        offx = x - wp->x;
        offy = y - wp->y;
#ifndef NANOGTK
        if (wp->unmapcount || !wp->output {
                if(wp->x != x || wp->y != y)
                        OffsetWindow(wp, offx, offy);
                wp->width = width;
                wp->height = height;
                return;
        }

#else
        if (wp->unmapcount || !wp->output || ( (wp->width==1)&&(wp->height==1))
                || (width + x <= 0) || (height + y <= 0 ) ) {
                if(wp->x != x || wp->y != y)
                        OffsetWindow(wp, offx, offy);
                wp->width = width;
                wp->height = height;
                return;
        }

        if (wp->x == x && wp->y == y)
        {
                if (wp->width != width || wp->height != height)
                {
                        GrNanoGTKResizeWindow(wp,width,height);
                        GsDeliverUpdateEvent(wp, GR_UPDATE_SIZE, wp->x, wp->y, width, height);
                }
                return;
        }
        else if ((wp->width==width) && (wp->height==height) )
        {
                GrMoveWindow(wid,x - wp->parent->x,y - wp->parent->y,0);
                return;
        }



#endif

        /*
         * This should be optimized to not require redrawing of the window
         * when possible.
         */
        GsWpUnmapWindow(wp, GR_TRUE);
        if(wp->x != x || wp->y != y)
                OffsetWindow(wp, offx, offy);
        wp->width = width;
        wp->height = height;
        GsWpMapWindow(wp, GR_TRUE);
        GsDeliverUpdateEvent(wp, GR_UPDATE_SIZE, x, y, width,height);
}

/*
 * GrSegments() draws multiple, unconnected lines. For each segment, draw
 * a line between (x1, y1) and (x2, y2) and in the order listed in the array of
 * GR_SEGMENT structures & does not perform joining at coincident endpoints.
 */
void
GrSegments(GR_DRAW_ID id, GR_GC_ID gc, GR_SIZE nsegs, GR_SEGMENT *segs)
{
        GR_DRAWABLE     *dp;
        int i;
        GR_GC           *gcp;           /* graphics context */
#ifdef LINE_ATTRIB_SUPPORT
        GR_POINT                corners[4];
        GR_COUNT                n_dashes;
        GR_DASH_SIZE    *dashes;
        GR_DASH_SIZE    default_dash_length = GR_DEFAULT_DASH_SIZE;
        GR_SEGMENT              *segments;      /* dash-line-style segments */
        GR_COUNT                n_segments;
#endif

        gcp = GsFindGC(gc);
        switch (GsPrepareDrawing(id, gc, &dp)) {
                case GR_DRAW_TYPE_WINDOW:
                case GR_DRAW_TYPE_PIXMAP:
                        if (gcp == NULL) {
                                fprintf(stderr, "%s: GsFindGC failed\n", __FUNCTION__);
                return;
                        }
#ifdef LINE_ATTRIB_SUPPORT
                        if (gcp->line_style == GR_LINE_STYLE_SOLID) {
#endif
                                if (gcp->line_width == 1) {
                                        for (i=0; i < nsegs; i++) {
                                                GdLine (dp->psd, dp->x + segs[i].x1,
                                                                dp->y + segs[i].y1,
                                                                dp->x + segs[i].x2,
                                                                dp->y + segs[i].y2, TRUE);
                                        }
                                }
#ifdef LINE_ATTRIB_SUPPORT
                                else if (gcp->line_width > 1) {
                                        for (i=0; i < nsegs; i++) {
                                        /*
                                         * get the 4 points of the rectangle,
                                         * and draw it as a filled poly
                                         */
                                                calc_thick_line_corners(
                                                        (GR_POINT){segs[i].x1, segs[i].y1},
                                                        (GR_POINT){segs[i].x2, segs[i].y2},
                                                        gcp->line_width, corners);
                                                draw_thick_line(dp, corners);
                                        }
                                }
                        }
                        else if (gcp->line_style == GR_LINE_STYLE_DASH) {
                                n_dashes = (gcp->n_dashes > 0)? gcp->n_dashes : 1;
                                dashes = (gcp->n_dashes > 0)? gcp->dash_list : &default_dash_length;
                                for (i=0; i < nsegs; i++) {
                                        calc_dash_line_coords ( segs[i].x1, segs[i].y1,

        segs[i].x2, segs[i].y2,

        dashes, n_dashes,

        &segments, &n_segments);
                                        draw_dash_line(dp, segments, n_segments, gcp->line_width);
                                        free(segments);
                                }
                        }
#endif

            break;
        }
}


void
GrSetEncoding(int encoding)
{
#if HAVE_CJK_SUPPORT
        set_cjk_encoding(encoding);
        GsDeliverLangEvent(encoding,get_cjk_default_font_size());
#endif
}

void
GrGetEncoding(int* encoding, GR_SIZE *width)
{
#if HAVE_CJK_SUPPORT
        *encoding = get_cjk_encoding();
        if (*encoding!=0)
                *width = get_cjk_default_font_size();
        else
                *width = 0;
#else
        *encoding = EN_ENCODE;
        *width = 0;
#endif
}

/*
 * GrLines() uses the components of the specified GC to draw npoints-1 lines
 * between each pair of points (point[i], point[i+1]) in the array of GR_POINT
 * structures. It draws the lines in the order listed in the array. The
 * lines join correctly at all intermediate points, and if the first and last
 * points coincide, the first and last lines also join correctly.
 * This function treats all coordinates as relative to the origin.
 *
 * will draw filled-poly if line with > 1
 */
void
GrLines(GR_DRAW_ID id, GR_GC_ID gc, GR_COUNT npoints, GR_POINT *points)
{
        GR_DRAWABLE     *dp;
        GR_COUNT        i, j;
        GR_GC           *gcp;           /* graphics context */
#ifdef LINE_ATTRIB_SUPPORT
        GR_POINT                corners[4];                     /* for thick lines */
        GR_POINT                prev_corners[4];        /* for joining thick lines */
        GR_POINT                join_points[4];         /* 4 joining points of 2
        intersecting lines */
        GR_POINT                first_seg_corners[4];
        GR_POINT                last_seg_corners[4];
        GR_SEGMENT              *segments = NULL;       /* for dash-style lines */
        GR_COUNT                n_segments;
        GR_DASH_SIZE    default_dash_length = GR_DEFAULT_DASH_SIZE;
#endif

        gcp = GsFindGC(gc);
        switch (GsPrepareDrawing(id, gc, &dp)) {
        case GR_DRAW_TYPE_WINDOW:
        case GR_DRAW_TYPE_PIXMAP:
        //FIXME : is there's a maximum points could be drawn at a time ?
                if (gcp == NULL) {
                        fprintf(stderr, "%s: GsFindGC failed\n", __FUNCTION__);
            return;
        }
#if LINE_ATTRIB_SUPPORT
                if (gcp->line_style == GR_LINE_STYLE_SOLID) {
#endif
                        if (gcp->line_width == 1) {
                                for(i=1; i < npoints; i++)
                                {
                                        GdLine(dp->psd, dp->x + points[i-1].x,
                                                dp->y +points[i-1].y, dp->x + points[i].x,
                                                dp->y +points[i].y, TRUE);
                                }
                        }
#ifdef LINE_ATTRIB_SUPPORT
                        else if (gcp->line_width > 1) {
                                for(i=1; i < npoints; i++) {
                                        calc_thick_line_corners (points[i-1], points[i],
                                                                gcp->line_width, corners);
                                        draw_thick_line(dp, corners);
                                        if (i > 1) {
                                        /*
                         * fill join area composed of 4 points that
                     * forms intersections of the 2 thick lines
                                     */
                                                find_join_points(
                                                        &(prev_corners[3]), &(prev_corners[0]),
                                                        &(prev_corners[2]), &(prev_corners[1]),
                                                        &(corners[3]), &(corners[0]),
                                                        &(corners[2]), &(corners[1]),
                                                        join_points);
                                                add_dp_offset(dp, join_points, 4);
                                                GdFillPoly(dp->psd, 4, join_points);
                                        }
                                        memcpy(prev_corners, corners, 4 * sizeof(GR_POINT));
                                }
                        }
                }
                else if (gcp->line_style == GR_LINE_STYLE_DASH) {
                        for (i = 1; i < npoints; i++) {
                                if (gcp->n_dashes > 0)
                                        calc_dash_line_coords (

points[i-1].x, points[i-1].y,

points[i].x, points[i].y,

gcp->dash_list, gcp->n_dashes,

&segments, &n_segments);
                                else
                                        calc_dash_line_coords (

points[i-1].x, points[i-1].y,

points[i].x, points[i].y,

&default_dash_length, 1,

&segments, &n_segments);

                                for (j = 0; j < n_segments; j++) {
                                        if ((j&1) == 0) {
                                                if (gcp->line_width == 1) {
                                                        GdLine (dp->psd, segments[j].x1,
                                                                        segments[j].y1, segments[j].x2,
                                                                        segments[j].y2, TRUE);
                                                }
                                                else if (gcp->line_width > 1) {
                                                        calc_thick_line_corners (
                                                                        (GR_POINT){segments[j].x1, segments[j].y1},
                                                                        (GR_POINT){segments[j].x2, segments[j].y2},
                                                                        gcp->line_width, corners);
                                                        draw_thick_line(dp, corners);
                                                        if (j == 0)
                                                                memcpy (first_seg_corners,
                                                                        corners, 4*sizeof(GR_POINT));
                                                }
                                        }
                                }
                                if (gcp->line_width > 1) {
                                        if (i > 1) {
                                                /* fill join area */
                                                find_join_points(
                                                        &(last_seg_corners[3]), &(last_seg_corners[0]),
                                                        &(last_seg_corners[2]), &(last_seg_corners[1]),
                                                        &(first_seg_corners[3]), &(first_seg_corners[0]),
                                                        &(first_seg_corners[2]), &(first_seg_corners[1]),
                                                        join_points);
                                                add_dp_offset(dp, join_points, 4);
                                                GdFillPoly(dp->psd, 4, join_points);
                                        }
                                        memcpy (last_seg_corners,
                                                first_seg_corners, 4*sizeof(GR_POINT));
                                }
                                if (segments != NULL)
                                        free(segments);
                        }
                }
#endif
        break;
        }
}

#ifdef CALIBRATION
#include "../drivers/mou_tp.h"
#endif

/* Add by Jason for Calibration */
void GrCalibration(int perform_calibration, int a, int b, int c, int d, int e, int f, int s)
{
#ifdef CALIBRATION
        extern int global_perform_calibration;
        extern TRANSFORMATION_COEFFICIENTS global_calibration;

        global_perform_calibration = perform_calibration;

        if(perform_calibration)
        {
                global_calibration.a = a;
                global_calibration.b = b;
                global_calibration.c = c;
                global_calibration.d = d;
                global_calibration.e = e;
                global_calibration.f = f;
                global_calibration.s = s;
        }
#endif
}

void
GrAreaAlpha(GR_DRAW_ID id, GR_GC_ID gc, GR_COORD x, GR_COORD y, GR_SIZE width,
        GR_SIZE height, void *pixels, int pixtype, short alpha_value)
{
        GR_DRAWABLE     *dp;

        switch (GsPrepareDrawing(id, gc, &dp)) {
                case GR_DRAW_TYPE_WINDOW:
                case GR_DRAW_TYPE_PIXMAP:
                        GdArea(dp->psd, dp->x + x, dp->y + y, width, height,
                                pixels, pixtype/*, alpha_value*/);
                        break;
        }
}

/* Add by Jason to invert color of specific area */
void GrInvertColorArea(GR_DRAW_ID id, GR_GC_ID gc, int type, GR_COORD x, GR_COORD y, GR_SIZE width, GR_SIZE height)
{
/*
        printf("Invert color id=%d type=%d, x=%d y=%d, width=%d, height=%d\n", id, type, x, y, width, height);
*/

//      GrSetGCForeground(gc, WHITE);
        GrSetGCForeground(gc, BLACK);
        GrSetGCMode(gc, GR_MODE_XOR);
        GrFillRect(id, gc, x, y, width, height);
        GrSetGCMode(gc, GR_MODE_SET);
}

/*
 * This function also clear the area and the child area if the
 * child window is intersect with this area, no expose area will
 * be genrated.
 */
void
GrClearChildArea(GR_WINDOW_ID wid, GR_COORD x, GR_COORD y,
        GR_SIZE width, GR_SIZE height)
{
        GR_WINDOW *wp;
        MWCLIPREGION *userregion;

        wp = GsPrepareWindow(wid);

        if(!wp)
                return;

       if (wp->unmapcount || !wp->output)
                return;

        /*
         * Reduce the arguments so that they actually lie within the window.
         */
        if (x < 0) {
                width += x;
                x = 0;
        }
        if (y < 0) {
                height += y;
                y = 0;
        }
        if (x + width > wp->width)
                width = wp->width - x;
        if (y + height > wp->height)
                height = wp->height - y;

        /*
         * Now see if the region is really in the window.  If not, then
         * do nothing.
         */
        if ((x >= wp->width) || (y >= wp->height) || (width <= 0) ||
                (height <= 0))
                        return;
#if 0
        GsExposeArea(wp, wp->x + x, wp->y + y, width, height, NULL);
        return;
#endif

        userregion =GdAllocRectRegion(x, y, width, height);
        /*
         * Draw the background of the window.
         * Invalidate the current graphics context since
         * we are changing the foreground color and mode.
         */
        GsSetClipWindow(wp, userregion, GR_INCLUDE_INFERIORS);
        curgcp = NULL;
        GdSetMode(GR_MODE_SET);
        GdSetForeground(GdFindColor(wp->background));
        GdFillRect(wp->psd, wp->x + x, wp->y + y, width, height);
}

/**
 * GrConvertSelection:
 * selection Specifies the selection atom.
 * target Specifies the target atom.
 * property Specifies the property name. You also can pass None .
 * requestor Specifies the requestor.
 * time Specifies the time.
 */
void
GrConvertSelection(GR_ATOM_ID atom, GR_ATOM_ID target, GR_ATOM_ID prop,
                GR_WINDOW_ID requestor, GR_TIME time)
{
        GR_WINDOW *wp, *swp;
        GR_SELECTION *selectionp;

        if(!(wp=GsFindWindow(requestor)))
                return; /* Bad Window */

        selectionp = GsAtomIntern(atom);
        if(!selectionp)
                return; /* Bad Atom */

        /* The selection has an owner, send selectionrequest event to the owner */
        if(selectionp->owner)
        {
                swp = GsFindWindow(selectionp->owner);
                if(!swp)
                {
                        selectionp->owner = GR_NONE;
                        goto destroy;
                }
                GsDeliverSelectionEvent(swp, GR_EVENT_TYPE_SELECTION_REQUEST,
                        requestor, selectionp->atom, target,
                        prop, selectionp->time);
        }
        else
        {
destroy:
                /* Selection have no owner : property set to GR_NONE */
                GsDeliverSelectionEvent(wp, GR_EVENT_TYPE_SELECTION_NOTIFY,
                        requestor, atom, target, GR_NONE, time);
        }
        /* update the selection information */
        selectionp->time = time;
}

/*
 * GrGetAtomName()
 * atom : atom for the property name you want returned
 * string : return name of the atom.
 * Function returns the name associated with the specified atom.
 */
void
GrGetAtomName(GR_ATOM_ID atom, GR_CHAR **string)
{
        GR_SELECTION *selectionp;

        selectionp = GsAtomIntern(atom);
        if(!selectionp)
                *string=NULL;
        else
                *string = strdup(selectionp->string);
}

/**
 * GrInternAtom:
 * atom_name : Specifies the name associated with the atom you want returned.
 * only_if_exists : False, atom is created if it does not exist, otherwise.
 *
 * Uppercase and lowercase matter; the strings ``thing'', ``Thing'', and
 * ``thinG'' all designate different atoms. The atom will remain defined
 * even after the client's connection closes.
 * It will become undefined only when the last connection to the server closes.
 */
GR_ATOM_ID
GrInternAtom(const GR_CHAR* atom_name, int only_if_exists)
{
        GR_SELECTION *selectionp;
        GR_ATOM_ID found=-1;

        selectionp = GsFindAtom(atom_name);
        if(!selectionp && only_if_exists)
                return GR_NONE;
        /* Create a new selection */
        if(!only_if_exists)
        {
                if(!selectionp)
                {
                        selectionp = GsNewSelection();
                        selectionp->string = strdup(atom_name);
                        selectionp->owner = GR_NONE;
                        selectionp->time = 0;
                }
        }
        found = selectionp->atom;
        return found;
}

/*
 * GrGetWindowProp()
 * wid : owner of the property
 * prop_id : the atom of the property wish to get
 * Function returns the actual type of the property; the actual format of
 * the property; the number of 8-bit, 16-bit, or 32-bit items transferred;
 * a pointer to the data actually returned.
 */
void
GrGetWindowProp(GR_WINDOW_ID wid, GR_ATOM_ID prop_id, GR_ATOM_ID *ret_type,
        int *ret_format, unsigned long *nitems_return, GR_CHAR **prop_return)
{
        GR_SELECTION *selectionp;
        GR_PROP_LIST *prop;

        if(!(selectionp = GsAtomIntern(prop_id) ))
                return; /* BAD_PROPERTY ATOM */
        if(!GsFindWindow(wid))
                return; /* BAD_WINDOW */

        /* Do we have the data of this property */
        prop = GsFindProp(selectionp->atom, wid);
        if(!prop)
        {
                *ret_type = GR_NONE;
                return; /*NO Such property */
        }

        *ret_type =  prop->textprop.encoding;
        *ret_format = prop->textprop.format;
        *nitems_return = prop->textprop.nitems;
        *prop_return = prop->textprop.value;
}

/*
 * GrChangeProp()
 * Function alters the property for the specified window and
 * causes the server to generate a GR_EVENT_TYPE_PROPERTY_NOTIFY
 * event on that window.
 * The data is always replaced.
 * If the property doesn't exits it will be created.
 */
void
GrChangeProp(GR_WINDOW_ID wid, GR_ATOM_ID prop_id, GR_ATOM_ID type,
        GR_SIZE format, GR_CHAR *data, GR_SIZE nitems)
{
        GR_SELECTION * selp;
        GR_TEXT_PROP text_prop;
        GR_PROP_LIST *prop;
        GR_WINDOW *wp;

        wp = GsFindWindow(wid);
        if(!wp)
                return; /* Bad Window */

        /* Find the property atom */
        selp = GsAtomIntern(prop_id);
        if(!selp)
                return; /* Bad Property ATOM */

        /* Find the associate property with this window */
        prop = GsFindProp(prop_id, wid);

        if(prop)
        {
                prop->textprop.format = format;
                prop->textprop.value = data;
                prop->textprop.nitems = nitems;
                prop->textprop.encoding = type;
                prop->prop_id = prop_id;
                if(prop->textprop.value ==NULL)
                {
                        prop->textprop.value = (GR_CHAR*)malloc(nitems + 1);
                }
                memcpy(prop->textprop.value, data, nitems);
                prop->textprop.value[nitems] =0;
        }
        else
        {
                /* Not Found : create a new one */
                text_prop.format = format;
                text_prop.value = data;
                text_prop.nitems = nitems+1;
                text_prop.encoding = type;
                text_prop.value = data;
                GsAddPropList(text_prop, prop_id, wid);
                listpropdata->textprop.value[nitems] = 0;
                listpropdata->textprop.nitems = nitems;
        }
        GsDeliverPropertyEvent(wp, prop_id, GsGetCurTimems(),
                                GR_PROPERTY_NEW_VALUE);
}

/*
 * GrDeleteProp()
 * Ffunction deletes the specified property only if the property
 * was defined on the specified window and causes the server to generate
 * a GR_EVENT_TYPE_PROPERTY_NOTIFY event on the window
 * unless the property does not exist.
 */
void
GrDeleteProp(GR_WINDOW_ID wid, GR_ATOM_ID prop_id)
{
        GR_WINDOW *wp;
        GR_SIZE bool;

        wp = GsFindWindow(wid);

        if(!wp)
                return;

        bool = GsDeleteProperty(wid, prop_id);

        if(bool)
                GsDeliverPropertyEvent(wp, prop_id,
                        GsGetCurTimems(), GR_PROPERTY_DELETE);
}

int
GrGetRegion(GR_REGION_ID region, MWCLIPREGION *rgn)
{
        GR_REGION       *regionp;
        int             ret_val;

        regionp =  GsFindRegion(region);
        if (regionp == NULL) {
                memset(rgn, 0, sizeof(MWCLIPREGION));
                return MWREGION_ERROR;
        }

        memcpy(rgn,regionp->rgn,sizeof(MWCLIPREGION));
        return ret_val;
}

void
GrSetGCMask(GR_GC_ID gc, GR_COORD x, GR_COORD y, GR_COORD width, GR_WINDOW_ID wid)
{
        GR_GC           *gcp;
        GR_PIXMAP       *pp;

        gcp = GsFindGC(gc);
        if (!gcp)
                return;
        pp = GsFindPixmap(wid);
        if (pp)
        {
                gcp->mask.x = x;
                gcp->mask.y = y;
                gcp->mask.width = width;
                gcp->mask.height = pp->height;
                gcp->mask.rowstride = pp->psd->linelen;
                gcp->mask.mask = pp->psd->addr;
                gcp->changed = GR_TRUE;
        }
        else
                gcp->mask.mask = NULL;
}

void
GrTilePixmap(GR_WINDOW_ID wid, GR_WINDOW_ID pid, GR_GC_ID gcid,
        GR_COORD dst_x, GR_COORD dst_y,
        GR_COORD src_x, GR_COORD src_y,
        GR_SIZE width, GR_SIZE height)
{
        GR_PIXMAP       *wp;
        GR_PIXMAP       *pm;
        GR_DRAW_TYPE    type;
        GR_DRAWABLE     *dp;
        int X = dst_x;
        int Y = dst_y;
        int aX0 = X - src_x;
        int aY0 = Y - src_y;
        int aX1 = X + width;
        int aY1 = Y + height;
        int w, h;
        int sx, sy, dx, dy;
        int x,y;
        int mask_x, mask_y;

        type = GsPrepareDrawing(wid, gcid, &dp);
        if (type == GR_DRAW_TYPE_NONE)
                return;
        wp = (GR_PIXMAP *)dp;
        pm = GsFindPixmap(pid);
        if (!wp || !pm) return;

#if 0
        /* This doesn't work yet. Could speed up tiling for small tiles */
        if (wp->psd!=pm->psd ... and small tiles only)
        {
                extern MWCLIPREGION *clipregion;
                MWRECT          rc;

                rc.left = dst_x;
                rc.top = dst_y;
                rc.right = rc.left + width;
                rc.bottom = rc.top + height;
                if (rc.left < 0)
                {
                        rc.right+=rc.left;
                        rc.left = 0;
                }
                if (rc.top < 0)
                {
                        rc.bottom+=rc.top;
                        rc.top = 0;
                }

                if (GdRectInRegion(clipregion, &rc) == MWRECT_ALLIN) {
                        linear12_tile(wp->psd, wp->x + dst_x, wp->y + dst_y, width, height, pm->psd, src_x, src_y);
                        return;
                }
        }
#endif

        if (gr_mask)
        {
                GR_MASK local_mask_copy = *gr_mask;
                int off_mask_x = gr_mask->x - dst_x;
                int off_mask_y = gr_mask->y - dst_y;

                mask_x = gr_mask->x;
                mask_y = gr_mask->y;
                for (y = aY0; y < aY1; y += pm->height) {
                        h = MWMIN(pm->height, aY1 - y);
                        gr_mask->x = mask_x;
                        for (x = aX0; x < aX1; x += pm->width) {
                                w = MWMIN(pm->width, aX1 - x);
                                dx = x;
                                dy = y;
                                sx = sy = 0;
                                if (x < X) {
                                        dx = X;
                                        sx = src_x;
                                }
                                if (y < Y) {
                                        dy = Y;
                                        sy = src_y;
                                }
                                if ((w > 0) && (h > 0)) {
                                        GdBlit(wp->psd, wp->x + dx, wp->y + dy, w, h,
                                                   pm->psd, sx, sy, MWROP_SRCCOPY);
                                }
                                gr_mask->x += gr_mask->width;
                        }
                        gr_mask->y += gr_mask->height;
                }
                *gr_mask = local_mask_copy;
        }
        else
        {
                for (y = aY0; y < aY1; y += pm->height) {
                        h = MWMIN(pm->height, aY1 - y);
                        for (x = aX0; x < aX1; x += pm->width) {
                                w = MWMIN(pm->width, aX1 - x);
                                dx = x;
                                dy = y;
                                sx = sy = 0;
                                if (x < X) {
                                        dx = X;
                                        sx = src_x;
                                }
                                if (y < Y) {
                                        dy = Y;
                                        sy = src_y;
                                }
                                if ((w > 0) && (h > 0)) {
                                        GdBlit(wp->psd, wp->x + dx, wp->y + dy, w, h,
                                                   pm->psd, sx, sy, MWROP_SRCCOPY);
                                }
                        }
                }
        }
}

/**
 * GrGetSelectionOwner:
 * @typelist: pointer used to return the list of available mime types
 * @Returns: the ID of the window which currently owns the selection, or 0
 *
 * Finds the window which currently owns the selection and returns its ID,
 * or 0 if no window currently owns the selection. A pointer to the list of
 * mime types the selection owner is capable of supplying is placed in the
 * pointer specified by the typelist argument. The typelist is null terminated,
 * and the fields are seperated by newline characters. It is the callers
 * responsibility to free the typelist string, as it is allocated dynamically.
 */
GR_WINDOW_ID
GrGetSelectionOwner_aux(GR_ATOM_ID atom)
{
        GR_SELECTION*   selectionp=NULL;

        selectionp = GsAtomIntern(atom);
        if(!selectionp)
                return BAD_ATOM;

        if(!GsFindWindow(selectionp->owner))
        {
                /* Owner not exists */
                selectionp->owner = GR_NONE;
                return GR_NONE;
        }
        return selectionp->owner;
}

#ifdef CLIPBOARD_SUPPORT
/**
 * GrSetSelectionOwner:
 * @typelist: list of mime types selection data can be supplied as
 * @owner: the ID of the window to set the selection owner to
 *
 * Sets the current selection (otherwise known as the clipboard) ownership
 * to the specified window. The typelist argument is a list of mime types which
 * the window is able to supply the data as. At least one type must be
 * specified (typically text/plain for plain ASCII text or text/uri-list for
 * a filename). Specifying an owner of 0 disowns the selection.
 *
 * The window which owns the current selection must be prepared to handle
 * SELECTION_LOST events (received when another window takes ownership of the
 * selection) and CLIENT_DATA_REQ events (received when a client wishes to
 * retreive the selection data).
 */
void
GrSetSelectionOwner_aux(GR_WINDOW_ID owner, GR_ATOM_ID atom, GR_TIME time)
{

        GR_WINDOW       *wp=NULL;
        GR_SELECTION    *selectionp;


        if(owner)
                wp = GsFindWindow(owner);
        if(!wp)
                owner = GR_NONE;

        if((selectionp = GsAtomIntern(atom)))
        {
                /* If new owner ( client or none) is not same as cur owner
                * send a SELECTION_CLEAR to the current owner
                * and current owner must not be none.
                */
                if(selectionp->owner && selectionp->owner!=owner)
                {
                        if((wp = GsFindWindow(selectionp->owner)))
                        {
                                GsDeliverSelectionEvent(wp,
                                GR_EVENT_TYPE_SELECTION_CLEAR,
                                wp->id, atom, GR_NONE, GR_NONE,
                                GsGetCurTimems());
                        }
                }
                selectionp->owner = owner;
                selectionp->time = time;
        }

}


/*
 * GrInjectSelectionEvent()
 * Generate a selection event and send to the given window.
 * The window will be deactivate/suspended if, the corresponding
 * client is suspended before -- this is only valid for EmLinux
 * desktop.
 */
void
GrInjectSelectionEvent(GR_EVENT_SELECTION *event)
{
        GR_WINDOW *wp=NULL;
        wp = GsFindWindow(event->owner);
        if(!wp)
        {
                wp = GsFindWindow(event->requestor);
                if(!wp)
                        return;
        }
                        GsDeliverSelectionEvent(wp, event->type, event->requestor,
                        event->selection, event->target, event->property,
                        event->time);
        GsDeactivateWin();
}

#endif // CLIPBOARD_SUPPORT

int GrGetFontList_aux(const char *name, char **list[])
{
	GR_FONTLIST **fonts;
	int i, numfonts;

	GdGetFontList(&fonts, &numfonts);

        if (numfonts)
        {
		*list = (char **)malloc((numfonts + 1) * sizeof(char *));

		for (i = 0; i < numfonts; i++)
		   (*list)[i] = strdup(fonts[i]->mwname);

        	GdFreeFontList(&fonts, numfonts);
        }

	DPRINTF("GrGetFontList_aux: numfonts=%d\n", numfonts);

	return numfonts;
}

#endif // srvfunc_gtkC_c

