#ifndef srvutil_gtkC_c
#define srvutil_gtkC_c

/*
 * GsFindTopLevelParent()
 * window : the top level window which we wish to find.
 * This function is design specific to EmLinux Desktop. Since
 * the client application will be suspended by the Task Manager,
 * to release the resource for the active client. Hence, we need to
 * find out which client are this window belong to , has it be suspended.
 * The criteria which determine this is to find out is the top level
 * parent have been unmap or not.
 */
GR_WINDOW*
GsFindTopLevelParent(GR_WINDOW *wp)
{
        GR_WINDOW       *pwp;
        pwp = wp->parent;

        while(pwp && pwp->parent != rootwp)
                pwp = pwp->parent;

        return pwp;
}

void GsExposeAreaOnlyChild(GR_WINDOW *wp, GR_COORD rootx, GR_COORD rooty, GR_SIZE width, GR_SIZE height)
{

        if (!wp->output || wp->unmapcount)
                return;

        /*
         * First see if the area overlaps the window including the border.
         * If not, then there is nothing more to do.
         */
        if ((rootx >= wp->x + wp->width + wp->bordersize) ||
                (rooty >= wp->y + wp->height + wp->bordersize) ||
                (rootx + width <= wp->x - wp->bordersize) ||
                (rooty + height <= wp->y - wp->bordersize))
                        return;

        /*
         * The area does overlap the window.  See if the area overlaps
         * the border, and if so, then redraw it.
         */
        if ((rootx < wp->x) || (rooty < wp->y) ||
                (rootx + width > wp->x + wp->width) ||
                (rooty + height > wp->y + wp->height))
                        if(wp->bordersize) GsDrawBorder(wp);
        {
                int l_x,l_y,l_width,l_height;

                l_x = rootx - wp->x;
                l_y = rooty - wp->y;
                l_width = width;
                l_height = height;
                if (l_x < 0) {
                        l_width += l_x;
                        l_x = 0;
                }
                if (l_y < 0) {
                        l_height += l_y;
                        l_y = 0;
                }
                if (l_x + l_width > wp->width)
                        l_width = wp->width - l_x;
                if (l_y + height > wp->height)
                        l_height = wp->height - l_y;

                curgcp = NULL;

                GsDeliverExposureEvent(wp, l_x, l_y, l_width, l_height);
        }
        /*
         * Now do the same for all the children.
         */
        for (wp = wp->children; wp; wp = wp->siblings)
                GsExposeArea(wp, rootx, rooty, width, height, NULL);
}

/*
 * This function will get the system current time and return
 * in mini second.
 */
GR_TIME
GsGetCurTimems()
{
        struct timeval tv;
        struct timezone tz;

        gettimeofday(&tv,&tz);

        return tv.tv_sec*1000+tv.tv_usec/1000;
}

pop_input_list *add_input(GR_WINDOW *wp, pop_input_list *l)
{
        pop_input_list * new_input;

        new_input = malloc(sizeof(pop_input_list));
        new_input->pop_win=wp;
        new_input->next=l;
        l = new_input;

        return l;
}

pop_input_list * remove_input(GR_WINDOW *wp, pop_input_list *l)
{
        pop_input_list * start_l=l;
        pop_input_list * cur_l;

        if (l->pop_win==wp)
        {
                start_l = l->next;
                free(l);
                return start_l;
        }
        while (l->next)
        {
                if (l->next->pop_win==wp)
                {
                        cur_l = l->next;
                        l->next = l->next->next;
                        free(cur_l);
                        return start_l;
                }
        }
        printf("WARNING : Can't find popup\n");
}

#endif // srvutil_gtkC_c

